from pydantic import BaseModel
from typing import Dict, Any, List, Optional


class NodeEvent(BaseModel):
    """
    node_type 就是 NodeType的值，但是不能直接使用 NodeType 因為會circular import
    """

    node_id: str
    node_type: str
    node_title: str
    is_print: bool = False


class NodeRunStartedEvent(NodeEvent):
    pass


class NodeRunCompletedEvent(NodeEvent):
    outputs: Dict[str, Any]
    complete_output: Optional[str] = None


class NodeRunStreamEvent(NodeEvent):
    chunk: str


class NodeRunFailedEvent(NodeEvent):
    error: str


class WorkflowRunStartedEvent(BaseModel):
    pass


class WorkflowRunCompletedEvent(BaseModel):
    outputs: Dict[str, Any]
    execution_times: List[str] = []


class WorkflowRunFailedEvent(BaseModel):
    error: str



import os
from dotenv import load_dotenv

load_dotenv()

TAIDE_MODELS = [
    "openai/Llama-3.1-405B-Instruct-FP8",
    "openai/Llama-3.1-70B",
    "openai/Llama-3.1-Nemotron-70B-Instruct",
    "openai/Llama3-TAIDE-LX-70B-Chat",
    "openai/TAIDE-LX-70B-Chat",
]


def get_api_key(model_name: str) -> str:
    if model_name in TAIDE_MODELS:
        return os.getenv("TAIDE_API_KEY", "")
    elif model_name.startswith("anthropic"):
        return os.getenv("ANTHROPIC_API_KEY", "")
    elif model_name.startswith("openai"):
        return os.getenv("OPENAI_API_KEY", "")
    elif model_name.startswith("gemini"):
        return os.getenv("GEMINI_API_KEY", "")
    elif model_name.startswith("together_ai"):
        return os.getenv("TOGETHERAI_API_KEY", "")
    elif model_name.startswith("deepinfra"):
        return os.getenv("DEEPINFRA_API_KEY", "")
    elif model_name.startswith("botrun"):
        return os.getenv("BOTRUN_API_KEY", "")
    else:
        return ""


def get_base_url(model_name: str) -> str:
    if model_name in TAIDE_MODELS:
        return os.getenv("TAIDE_BASE_URL", "")
    elif model_name.startswith("botrun"):
        return os.getenv("BOTRUN_BASE_URL", "")
    else:
        return ""


def get_custom_llm_provider(model_name: str) -> str:
    if model_name.startswith("botrun"):
        return "openai"
    else:
        return None



from typing import Union, List, Tuple
from google.cloud.exceptions import GoogleCloudError
from botrun_flow_lang.constants import USER_WORKFLOW_STORE_NAME
from botrun_flow_lang.services.user_workflow.user_workflow import UserWorkflow
from botrun_flow_lang.services.base.firestore_base import FirestoreBase
from google.cloud import firestore


class UserWorkflowFsStore(FirestoreBase):
    def __init__(self, env_name: str):
        super().__init__(f"{env_name}-{USER_WORKFLOW_STORE_NAME}")

    async def get_user_workflow(self, workflow_id: str) -> Union[UserWorkflow, None]:
        doc_ref = self.collection.document(workflow_id)
        doc = doc_ref.get()
        if doc.exists:
            data = doc.to_dict()
            return UserWorkflow(**data)
        else:
            print(f">============Getting user workflow {workflow_id} not exists")
            return None

    async def set_user_workflow(self, user_workflow: UserWorkflow):
        try:
            doc_ref = self.collection.document(user_workflow.id)
            doc_ref.set(user_workflow.model_dump())
            return True, user_workflow
        except GoogleCloudError as e:
            print(f"Error setting user workflow {user_workflow.id}: {e}")
            return False, None

    async def delete_user_workflow(self, workflow_id: str):
        try:
            doc_ref = self.collection.document(workflow_id)
            doc_ref.delete()
            return True
        except GoogleCloudError as e:
            print(f"Error deleting user workflow {workflow_id}: {e}")
            return False

    async def get_user_workflows(
        self, user_id: str, offset: int = 0, limit: int = 20
    ) -> Tuple[List[UserWorkflow], str]:
        try:
            query = (
                self.collection.where(
                    filter=firestore.FieldFilter("user_id", "==", user_id)
                )
                .offset(offset)
                .limit(limit)
            )

            docs = query.stream()
            workflows = [UserWorkflow(**doc.to_dict()) for doc in docs]
            return workflows, ""
        except GoogleCloudError as e:
            print(f"Error getting user workflows for user {user_id}: {e}")
            return [], str(e)



from pydantic import BaseModel


class UserSetting(BaseModel):
    user_id: str
    default_model: str = ""
    audio_reply: bool = False
    search_vendor: str = "Botrun"
    search_enabled: bool = False



import asyncio
from copy import deepcopy
import sys
import time
from pathlib import Path
from typing import Dict, List
from botrun_flow_lang.models.botrun_app import BotrunApp, BotrunAppMode
from botrun_flow_lang.models.nodes.answer_node import AnswerNodeData
from botrun_flow_lang.models.nodes.base_node import NodeType
from botrun_flow_lang.models.nodes.code_node import CodeNodeData
from botrun_flow_lang.models.nodes.event import (
    NodeRunCompletedEvent,
    NodeRunStartedEvent,
    NodeRunStreamEvent,
    WorkflowRunCompletedEvent,
    WorkflowRunFailedEvent,
)
from botrun_flow_lang.models.nodes.http_request_node import (
    Body,
    BodyType,
    HttpMethod,
    HttpRequestNodeData,
)
from botrun_flow_lang.models.nodes.iteration_node import IterationNodeData
from botrun_flow_lang.models.nodes.perplexity_node import (
    PerplexityModelConfig,
    PerplexityNodeData,
)
from botrun_flow_lang.models.nodes.search_and_scrape_node import (
    SearchAndScrapeNode,
    SearchAndScrapeNodeData,
)
from botrun_flow_lang.models.nodes.vertex_ai_search_node import VertexAiSearchNodeData
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.workflow import WorkflowData, Workflow
from botrun_flow_lang.models.nodes.start_node import StartNode, StartNodeData
from botrun_flow_lang.models.nodes.llm_node import LLMNodeData, LLMModelConfig
from botrun_flow_lang.models.nodes.end_node import EndNodeData
from botrun_flow_lang.api.workflow.workflow_engine import run_workflow
from botrun_flow_lang.models.workflow_config import WorkflowConfig
from dotenv import load_dotenv
import logging
from datetime import datetime

from botrun_flow_lang.services.user_workflow.user_workflow import UserWorkflow

load_dotenv()


def get_llm_workflow_config():
    botrun_app = BotrunApp(
        name="波文件問",
        description="給波文件問答用的app",
        mode=BotrunAppMode.CHATBOT,
    )

    start_node = StartNodeData(
        title="Start",
    )

    model_config = LLMModelConfig(
        completion_params={
            "max_tokens": 4096,
            "temperature": 0.7,
        },
        mode="chat",
        name="gpt-4o-2024-08-06",
        provider="openai",
    )
    llm_node = LLMNodeData(
        title="LLM",
        model=model_config,
        prompt_template=[
            {
                "role": "system",
                "content": "妳是臺灣人，回答要用臺灣繁體中文正式用語，需要的時候也可以用英文，可以親切、俏皮、幽默，但不能隨便輕浮。在使用者合理的要求下請盡量配合他的需求，不要隨便拒絕",
            },
            {
                "role": "user",
                "content": f"{{{{#{start_node.id}.user_input#}}}}",
            },
        ],
        input_variables=[
            InputVariable(node_id=start_node.id, variable_name="user_input")
        ],
        output_variables=[
            OutputVariable(variable_name="llm_output"),
        ],
    )
    answer_node = AnswerNodeData(
        title="Answer",
        input_variables=[
            InputVariable(node_id=llm_node.id, variable_name="llm_output")
        ],
    )
    workflow = WorkflowData(nodes=[start_node, llm_node, answer_node])
    original_workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
    yaml_str = original_workflow_config.to_yaml()
    return yaml_str


def get_subsidy_workflow_config():
    question_count = 20
    generate_questions_prompt = """
      請以臺灣人的角度,用繁體中文生成 {question_count} 個關於「{subject}」的最常見問題。
  這些問題應該:
  1. 涵蓋不同年齡層父母的疑慮
  2. 包含申請流程、資格條件、金額計算等實際問題
  3. 考慮到不同家庭情況(如單親、低收入戶等)
  4. 包括政策變更、地區差異等相關問題
  5. 涉及實際領取和使用津貼的相關疑問

  請確保問題精確、具體,並且反映臺灣民眾對「{subject}」的真實需求和關切。
  請使用以下 JSON 格式嚴格回應,只包含問題內容:
  [
    {{"問題": "第1個問題內容"}},
    {{"問題": "第2個問題內容"}},
    ...
    {{"問題": "第{question_count}個題內容"}}
    ]
"""

    generate_keywords_prompt = """
  請根據以下關於臺灣育兒津貼政策的問題,生成相關的關鍵字群。每個問題對應一個關鍵字群,關鍵字群應包含3-5個重要的關鍵詞。
  這些關鍵字應該:
  1. 反映問題的核心主題
  2. 包含可能用於搜索相關資訊的詞彙
  3. 涵蓋問題中提到的重要概念或實體

  問題列表:
  {question_list}

  請使用以下 JSON 格式嚴格回應,只包含關鍵字群:
  [
    {{"關鍵字群": ["關鍵字1", "關鍵字2", "關鍵字3"]}},
    {{"關鍵字群": ["關鍵字1", "關鍵字2", "關鍵字3", "關鍵字4"]}},
    ...
  ]
"""
    model_config = LLMModelConfig(
        completion_params={},
        mode="chat",
        name="claude-3-5-sonnet-20241022",
        provider="anthropic",
    )
    botrun_app = BotrunApp(
        name="波津貼", description="有上網功能的 chatbot", mode=BotrunAppMode.CHATBOT
    )

    start_node = StartNodeData(
        title="Start",
    )

    generate_questions_by_llm_node = LLMNodeData(
        title="Generate Questions By LLM",
        model=model_config,
        prompt_template=[
            {
                "role": "user",
                "content": generate_questions_prompt.format(
                    question_count=question_count,
                    subject=f"{{{{#{start_node.id}.user_input#}}}}",
                ),
            },
        ],
        output_variables=[
            OutputVariable(variable_name="llm_output"),
        ],
    )
    get_question_list_from_json_node = CodeNodeData(
        title="Get Question List From JSON",
        code="""
import json
def main(llm_output):
    json_data = json.loads(llm_output)
    question_list = [item["問題"] for item in json_data]
    question_list= "\\n".join(question_list)
    return {"question_list": question_list}

        """,
        input_variables=[
            InputVariable(
                node_id=generate_questions_by_llm_node.id, variable_name="llm_output"
            )
        ],
        output_variables=[
            OutputVariable(variable_name="question_list"),
        ],
    )
    generate_keywords_llm_node = LLMNodeData(
        title="LLM",
        model=model_config,
        prompt_template=[
            {
                "role": "user",
                "content": generate_keywords_prompt.format(
                    question_list=f"{{{{#{get_question_list_from_json_node.id}.question_list#}}}}",
                ),
            },
        ],
        output_variables=[
            OutputVariable(variable_name="llm_output"),
        ],
    )
    get_keyword_list_from_json_node = CodeNodeData(
        title="Get Keyword List From JSON",
        code="""
import json

def main(llm_output):
    json_data = json.loads(llm_output)
    
    all_keywords = []
    for item in json_data:
        all_keywords.append(item["關鍵字群"])
    
    return {"keyword_list": " ".join(all_keywords[0])}
    """,
        input_variables=[
            InputVariable(
                node_id=generate_keywords_llm_node.id, variable_name="llm_output"
            )
        ],
        output_variables=[
            OutputVariable(variable_name="keyword_list"),
        ],
    )
    search_keywords_node = HttpRequestNodeData(
        title="Search Keywords",
        url="https://botrun-flow-lang-fastapi-prod-36186877499.asia-east1.run.app/api/search",
        method=HttpMethod.POST,
        body=Body(
            data={
                "query": f"{{{{#{get_keyword_list_from_json_node.id}.keyword_list#}}}}",
                "num": 3,
            },
            type=BodyType.JSON,
        ),
    )
    get_search_result_from_json_node = CodeNodeData(
        title="Get Search Result From JSON",
        code="""
import json

def main(body, keyword_list):
    search_results = json.loads(body)
    all_search_results = {}
    if "items" in search_results:
        all_search_results[keyword_list] = search_results
    else:
        all_search_results["error"] = "No search results found"

    # 輸出
    formatted_results = json.dumps(all_search_results, ensure_ascii=False, indent=2)
    return {"search_results": formatted_results}

    """,
        input_variables=[
            InputVariable(node_id=search_keywords_node.id, variable_name="body"),
            InputVariable(
                node_id=get_keyword_list_from_json_node.id, variable_name="keyword_list"
            ),
        ],
        output_variables=[
            OutputVariable(variable_name="search_results"),
        ],
    )
    answer_node = AnswerNodeData(
        title="Answer",
        input_variables=[
            InputVariable(
                node_id=generate_keywords_llm_node.id, variable_name="llm_output"
            )
        ],
    )
    workflow = WorkflowData(
        nodes=[
            start_node,
            generate_questions_by_llm_node,
            get_question_list_from_json_node,
            generate_keywords_llm_node,
            get_keyword_list_from_json_node,
            search_keywords_node,
            get_search_result_from_json_node,
            answer_node,
        ]
    )
    original_workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
    yaml_str = original_workflow_config.to_yaml()
    return yaml_str


def save_yaml_to_file(yaml_str: str, filename: str):
    templates_dir = Path(__file__).parent / "templates"
    templates_dir.mkdir(exist_ok=True)
    file_path = templates_dir / filename
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(yaml_str)
    print(f"YAML 已保存到: {file_path}")


def load_yaml_from_file(filename: str) -> str:
    templates_dir = Path(__file__).parent / "templates"
    file_path = templates_dir / filename
    with open(file_path, "r", encoding="utf-8") as f:
        return f.read()


def get_perplexity_like_workflow_config():
    is_async = True
    generate_questions_prompt = """
    你是一個專業的調查員，你會依據以下問題，去網路上搜尋相關資料，並且回答使用者。
    當使用者輸入一個問題時，你會
    1. 理解查詢：理解用戶輸入的查詢。這不僅僅是簡單的關鍵字匹配，而是深入分析查詢的上下文和意圖，以便更準確地理解用戶需求。
    2. 重構查詢：在理解查詢後，你會重構查詢以適應其搜索和分析模型。這包括將用戶的自然語言問題轉換為可以在網路上有效搜索的訊息格式，從而提高搜索效率和結果的相關性。
    3. 條列重構查詢：將重構後的查詢，條列成多個子問題，每個子問題都是一個可以在網路上搜尋到的具體問題。

    以下是使用者輸入的問題:
    {question}

    請使用以下 JSON 格式嚴格回應,只包含問題內容:
    [
        "第1個子問題",
        "第2個子問題",
        "最後一個子問題"
    ]
"""
    analyze_search_result_prompt = """請分析以下 Google 搜尋結果，參考 snippet 的內容，選出最相關的三個網頁連結。
問題: {question}
搜尋結果: {search_results}

請使用以下 JSON 格式嚴格回應:
{{
    "selected_urls": ["url1", "url2", "url3"]
}}"""
    note_taker_prompt = """你是一位資料記錄者，請分析以下網頁內容是否與問題相關，並做出詳實記錄。

用來 Google 搜尋的問題: {question}
網頁URL: {url}
網頁標題: {title}
網頁內容: {markdown}

如果內容相關，請提取相關的重要資訊並做詳實記錄。
請使用以下 JSON 格式嚴格回應，不要附加任何其它文字:
{{
    "url": "網頁URL",
    "title": "網頁標題",
    "note": "詳實的記錄內容"
}}

如果內容不相關，請回傳空值。
請使用以下 JSON 格式嚴格回應，不要附加任何其它文字:
{{
    "url": "",
    "title": "",
    "note": ""
}}
"""
    answer_prompt = """
    你是一個專業的資訊分析員，你會依據以下使用者的問題，以及網路搜尋到的資訊，統整出最完整的回答。
    如果你沒有獲得資訊，請誠實的告訴使用者，不要自行回答。

    {all_info}
    """
    claude_model_config = LLMModelConfig(
        completion_params={},
        name="anthropic/claude-3-5-sonnet-20241022",
    )
    openai_mini_model_config = LLMModelConfig(
        completion_params={},
        name="openai/gpt-4o-mini",
    )
    groq_model_config = LLMModelConfig(
        completion_params={},
        name="groq/llama-3.1-70b-versatile",
    )
    final_model_config = LLMModelConfig(
        completion_params={},
        # name="anthropic/claude-3-5-sonnet-20241022",
        name="gemini/gemini-1.5-pro",
    )
    botrun_app = BotrunApp(
        name="Perplexity Like Search",
        description="有上網搜尋功能的 chatbot",
        mode=BotrunAppMode.CHATBOT,
    )

    start_node = StartNodeData(
        title="Start",
    )
    search_and_scrape_node = SearchAndScrapeNodeData(
        title="Search And Scrape",
        search_query=f"{{{{#{start_node.id}.user_input#}}}}",
        print_stream=True,
    )
    generate_questions_by_llm_node = LLMNodeData(
        title="產生相關議題的問題集...",
        model=openai_mini_model_config,
        prompt_template=[
            {
                "role": "user",
                "content": generate_questions_prompt.format(
                    question=f"{{{{#{start_node.id}.user_input#}}}}",
                ),
            },
        ],
        input_variables=[
            InputVariable(node_id=start_node.id, variable_name="history"),
        ],
        print_start=True,
    )
    split_question_list_node = CodeNodeData(
        title="Split Question List",
        code="""
import json
def main(llm_output):
    question_list = json.loads(llm_output)
    return {"question_list": question_list}
        """,
        input_variables=[
            InputVariable(
                node_id=generate_questions_by_llm_node.id, variable_name="llm_output"
            )
        ],
        output_variables=[
            OutputVariable(variable_name="question_list"),
        ],
        # print_complete=True,
    )
    display_doing_research_node = CodeNodeData(
        title="將產生的問題進行 Google 搜尋...",
        code="""
def main():
    return {
        "result": ""
    }
        """,
        input_variables=[],
        output_variables=[],
        print_start=True,
    )
    iteration_question_node = IterationNodeData(
        title="Iteration",
        input_selector=InputVariable(
            node_id=split_question_list_node.id, variable_name="question_list"
        ),
        output_selector=InputVariable(node_id="", variable_name=""),
        is_async=is_async,
    )
    search_question_node = HttpRequestNodeData(
        title="將問題進行 Google 搜尋...",
        url="https://botrun-flow-lang-fastapi-dev-36186877499.asia-east1.run.app/api/search",
        method=HttpMethod.POST,
        body=Body(
            data={
                "query": f"{{{{#{iteration_question_node.id}.item#}}}}",
                "num": 10,
            },
            type=BodyType.JSON,
        ),
    )

    analyze_search_results_node = LLMNodeData(
        title="從搜尋結果中選出最相關的網頁連結...",
        model=openai_mini_model_config,
        prompt_template=[
            {
                "role": "user",
                "content": analyze_search_result_prompt.format(
                    question=f"{{{{#{iteration_question_node.id}.item#}}}}",
                    search_results=f"{{{{#{search_question_node.id}.body#}}}}",
                ),
            },
        ],
        input_variables=[],
        stream=False,
        print_stream=False,
    )

    get_selected_urls_node = CodeNodeData(
        title="取得選定的 URL...",
        code="""
import json
from urllib.parse import quote
def main(llm_output):
    try:
        result = json.loads(llm_output)
        urls = result.get("selected_urls", [])
        if not urls:
            print("No URLs found in LLM output")
            return {"links": []}
            
        final_urls = [quote(url, safe=':/') for url in urls]
        print("Get Selected URLs....")
        print(final_urls)
        return {"links": final_urls}
    except json.JSONDecodeError as e:
        print(f"Error parsing LLM output: {str(e)}")
        return {"links": []}
    except Exception as e:
        print(f"Unexpected error: {str(e)}")
        return {"links": []}
        """,
        input_variables=[
            InputVariable(
                node_id=analyze_search_results_node.id, variable_name="llm_output"
            ),
        ],
        output_variables=[
            OutputVariable(variable_name="links"),
        ],
        # print_start=True,
        # print_complete=True,
    )

    iteration_scrape_all_pages_node = IterationNodeData(
        title="Iteration Scrape All Pages",
        input_selector=InputVariable(
            node_id=get_selected_urls_node.id, variable_name="links"
        ),
        output_selector=InputVariable(node_id="", variable_name=""),
        is_async=is_async,
    )
    scrape_page_node = HttpRequestNodeData(
        title="Scraping Web Page",
        url="https://botrun-crawler-fastapi-prod-36186877499.asia-east1.run.app/scrape",
        method=HttpMethod.GET,
        params=f"url: {{{{#{iteration_scrape_all_pages_node.id}.item#}}}}",
        # print_start=True,
        # print_complete=True,
    )

    get_scrape_result_markdown_node = CodeNodeData(
        title="將網頁內容轉換成 Markdown 格式...",
        code="""
import json
def main(body, status_code):
    try:
        if status_code == 200:
            body = json.loads(body)
            return {
                "markdown": body["data"]["markdown"],
                "title": body["data"]["metadata"]["title"],
            }
        else:
            return {"markdown": "", "title":""}
    except Exception as e:
        print(f"Error processing [將網頁內容轉換成 Markdown 格式...] result: {str(e)}")
        return {"markdown": "", "title": ""}
        """,
        input_variables=[
            InputVariable(node_id=scrape_page_node.id, variable_name="body"),
            InputVariable(node_id=scrape_page_node.id, variable_name="status_code"),
        ],
        output_variables=[
            OutputVariable(variable_name="markdown"),
            OutputVariable(variable_name="title"),
        ],
        # print_start=True,
    )
    analyze_scrape_result_node = LLMNodeData(
        title="分析網頁內容...",
        model=openai_mini_model_config,
        prompt_template=[
            {
                "role": "user",
                "content": note_taker_prompt.format(
                    question=f"{{{{#{iteration_question_node.id}.item#}}}}",
                    url=f"{{{{#{iteration_scrape_all_pages_node.id}.item#}}}}",
                    title=f"{{{{#{get_scrape_result_markdown_node.id}.title#}}}}",
                    markdown=f"{{{{#{get_scrape_result_markdown_node.id}.markdown#}}}}",
                ),
            },
        ],
        input_variables=[],
        stream=False,
        print_stream=False,
    )
    iteration_scrape_all_pages_node.output_selector = InputVariable(
        node_id=analyze_scrape_result_node.id, variable_name="llm_output"
    )

    combine_all_scrape_results_node = CodeNodeData(
        title="Combine All Scrape Results",
        code="""
import json
def main(item, output):
    results=[]
    for result in output:
        data = json.loads(result)
        if data.get("url") and data.get("title") and data.get("note"):
            data["google_question"] = item
            results.append(data)
    return {"results": results}
        """,
        input_variables=[
            InputVariable(node_id=iteration_question_node.id, variable_name="item"),
            InputVariable(
                node_id=iteration_scrape_all_pages_node.id, variable_name="output"
            ),
        ],
        output_variables=[
            OutputVariable(variable_name="results"),
        ],
        # print_start=True,
        # print_complete=True,
    )
    iteration_question_node.output_selector = InputVariable(
        node_id=combine_all_scrape_results_node.id, variable_name="results"
    )

    consolidate_all_info = CodeNodeData(
        title="Consolidating All Information",
        code="""
import json
def main(user_input, results):
    all_info = f"使用者輸入的問題:\\n{user_input}\\n\\n"
    data = json.dumps(results, ensure_ascii=False, indent=2)
    # print(daa[:200])
    all_info += f"網路搜尋回來的資訊:\\n{data}\\n"
        
    return {"all_info": all_info}
""",
        input_variables=[
            InputVariable(node_id=start_node.id, variable_name="user_input"),
            InputVariable(node_id=search_and_scrape_node.id, variable_name="results"),
        ],
        output_variables=[
            OutputVariable(variable_name="all_info"),
        ],
        # print_start=True,
        # print_complete=True,
    )

    answer_llm_node = LLMNodeData(
        title="做完研究後最後的回答...",
        model=claude_model_config,
        prompt_template=[
            {
                "role": "user",
                "content": answer_prompt.format(
                    all_info=f"{{{{#{consolidate_all_info.id}.all_info#}}}}",
                ),
            },
        ],
        input_variables=[InputVariable(node_id=start_node.id, variable_name="history")],
        print_start=True,
    )
    get_reference_links_node = CodeNodeData(
        title="取得參考資料的連結...",
        code="""
import json
def main(results):
    references = []
    for result in results:
        if result.get("note") and result.get("url") and result.get("title"):
            references.append(f"- [{result['title']}]({result['url']})")
    
    if references:
        references=list(set(references))
        text = "參考資料:\\n" + "\\n".join(references)
        return {"result": text}
    return {"result": ""}
""",
        input_variables=[
            InputVariable(node_id=search_and_scrape_node.id, variable_name="results")
        ],
        output_variables=[
            OutputVariable(variable_name="result"),
        ],
        complete_output="result",
    )

    answer_node = AnswerNodeData(
        title="Answer",
        input_variables=[
            InputVariable(node_id=answer_llm_node.id, variable_name="llm_output")
        ],
    )
    workflow = WorkflowData(
        nodes=[
            start_node,
            search_and_scrape_node,
            # generate_questions_by_llm_node,
            # split_question_list_node,
            # display_doing_research_node,
            # [
            #     iteration_question_node,
            #     search_question_node,
            #     analyze_search_results_node,
            #     get_selected_urls_node,
            #     [
            #         iteration_scrape_all_pages_node,
            #         scrape_page_node,
            #         get_scrape_result_markdown_node,
            #         analyze_scrape_result_node,
            #     ],
            #     combine_all_scrape_results_node,
            # ],
            consolidate_all_info,
            answer_llm_node,
            get_reference_links_node,
            answer_node,
        ]
    )
    original_workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
    return original_workflow_config


def get_perplexity_workflow_config():
    botrun_app = BotrunApp(
        name="Perplexity Search",
        description="使用 perplexity 搜尋的 chatbot",
        mode=BotrunAppMode.CHATBOT,
    )
    start_node = StartNodeData(
        title="Start",
    )
    perplexitly_model_config = PerplexityModelConfig(
        completion_params={},
        name="llama-3.1-sonar-huge-128k-online",
    )

    perplexity_node = PerplexityNodeData(
        title="Perplexity Search",
        model=perplexitly_model_config,
        prompt_template=[
            {
                "role": "user",
                "content": f"""
                幫我根據使用者的提問進行回答，你會尋找最新的資料，而且會以政府網站為優先，最後加上參考資料來源。

                以下是使用者的提問：
                {{{{#{start_node.id}.user_input#}}}}""",
            },
        ],
        input_variables=[
            InputVariable(node_id=start_node.id, variable_name="history"),
        ],
    )
    answer_node = AnswerNodeData(
        title="Answer",
        input_variables=[
            InputVariable(node_id=perplexity_node.id, variable_name="perplexity_output")
        ],
    )
    workflow = WorkflowData(
        nodes=[
            start_node,
            perplexity_node,
            answer_node,
        ]
    )
    original_workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
    return original_workflow_config


def get_vertexai_workflow_config():
    answer_prompt = """
    你是一個專業的資訊分析員，你會依據以下使用者的問題，以及網路搜尋到的資訊，統整出最完整的回答。
    如果你沒有獲得資訊，請誠實的告訴使用者，不要自行回答。

    {all_info}
    """
    claude_model_config = LLMModelConfig(
        completion_params={},
        name="anthropic/claude-3-5-sonnet-20241022",
    )
    gemini_model_config = LLMModelConfig(
        completion_params={},
        name="gemini/gemini-1.5-pro",
    )
    botrun_app = BotrunApp(
        name="Vertex AI Search",
        description="使用 Vertex AI 搜尋的 chatbot",
        mode=BotrunAppMode.CHATBOT,
    )
    start_node = StartNodeData(
        title="Start",
    )

    vertexai_search_node = VertexAiSearchNodeData(
        title="Vertex AI Search",
        search_query=f"{{{{#{start_node.id}.user_input#}}}}",
        project_id="scoop-386004",
        location="global",
        data_store_ids=["tw-gov-welfare_1730944342934"],
        print_stream=True,
    )
    consolidate_all_info = CodeNodeData(
        title="Consolidating All Information",
        code="""
import json
def main(user_input, results):
    all_info = f"使用者輸入的問題:\\n{user_input}\\n\\n"
    data = json.dumps(results, ensure_ascii=False, indent=2)
    # print(daa[:200])
    all_info += f"網路搜尋回來的資訊:\\n{data}\\n"
        
    return {"all_info": all_info}
""",
        input_variables=[
            InputVariable(node_id=start_node.id, variable_name="user_input"),
            InputVariable(node_id=vertexai_search_node.id, variable_name="results"),
        ],
        output_variables=[
            OutputVariable(variable_name="all_info"),
        ],
        # print_start=True,
        # print_complete=True,
    )

    answer_llm_node = LLMNodeData(
        title="做完研究後最後的回答...",
        model=gemini_model_config,
        prompt_template=[
            {
                "role": "user",
                "content": answer_prompt.format(
                    all_info=f"{{{{#{consolidate_all_info.id}.all_info#}}}}",
                ),
            },
        ],
        input_variables=[InputVariable(node_id=start_node.id, variable_name="history")],
        print_start=True,
    )
    get_reference_links_node = CodeNodeData(
        title="取得參考資料的連結...",
        code="""
import json
def main(results):
    references = []
    for result in results:
        if result.get("note") and result.get("url") and result.get("title"):
            references.append(f"- [{result['title']}]({result['url']})")
    
    if references:
        references=list(set(references))
        text = "參考資料:\\n" + "\\n".join(references)
        return {"result": text}
    return {"result": ""}
""",
        input_variables=[
            InputVariable(node_id=vertexai_search_node.id, variable_name="results")
        ],
        output_variables=[
            OutputVariable(variable_name="result"),
        ],
        complete_output="result",
    )

    answer_node = AnswerNodeData(
        title="Answer",
        input_variables=[
            InputVariable(node_id=answer_llm_node.id, variable_name="llm_output")
        ],
    )
    workflow = WorkflowData(
        nodes=[
            start_node,
            vertexai_search_node,
            consolidate_all_info,
            answer_llm_node,
            get_reference_links_node,
            answer_node,
        ]
    )
    original_workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
    return original_workflow_config


async def run_workflow_from_yaml(yaml_str, user_input, history: List[Dict] = []):
    # Setup logger
    log_dir = Path(__file__).parent / "log"
    log_dir.mkdir(exist_ok=True)

    timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    # log_file = log_dir / f"workflow-{timestamp}.log"

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
        handlers=[
            # logging.FileHandler(log_file, encoding="utf-8"),
            logging.StreamHandler(),
        ],
    )
    logger = logging.getLogger(__name__)

    logger.info(f"Starting workflow with user input: {user_input}")

    workflow_config = WorkflowConfig.from_yaml(yaml_str)
    workflow = Workflow.from_workflow_data(workflow_config.workflow)

    initial_variable_pool = {}
    # Set StartNode's user_input
    for item in workflow.items:
        if isinstance(item.node, StartNode):
            initial_variable_pool[item.node.data.id] = {}
            for output_variable in item.node.data.output_variables:
                if output_variable.variable_name == "history":
                    initial_variable_pool[item.node.data.id][
                        output_variable.variable_name
                    ] = history
                elif output_variable.variable_name == "user_input":
                    initial_variable_pool[item.node.data.id][
                        output_variable.variable_name
                    ] = user_input
            logger.info(f"Initialized StartNode with input: {user_input}")
            break

    generator = run_workflow(workflow, initial_variable_pool)
    async for event in generator:
        if isinstance(event, NodeRunStartedEvent) and event.is_print:
            # print(f"Node {event.node_title} started <=====")
            logger.info(f"Node {event.node_title} started <=====")
        elif isinstance(event, NodeRunStreamEvent) and event.is_print:
            print(event.chunk, end="", flush=True)
            # logger.info(event.chunk)
        elif isinstance(event, NodeRunCompletedEvent):
            if event.complete_output:
                content = event.outputs.get(event.complete_output, "")
                print(content)
            if event.is_print:
                logger.info(f"{event.outputs}")
        elif isinstance(event, WorkflowRunCompletedEvent):
            exe_times = deepcopy(event.execution_times)
            exe_times.sort(reverse=True)
            for exe_time in exe_times:
                print(exe_time)
            return event.outputs
        elif isinstance(event, WorkflowRunFailedEvent):
            error_msg = f"Workflow failed: {event.error}"
            print(f"\n{error_msg}")
            logger.error(error_msg)
            return None

    return None


if __name__ == "__main__":
    # yaml_str = get_llm_workflow_config()
    # final_output = asyncio.run(
    #     run_workflow_from_yaml(yaml_str, "告訴我一個小紅帽的故事")
    # )
    # exit(0)
    # yaml_str = get_subsidy_workflow_config()
    # yaml_file_name = "subsidy_workflow.yml"
    # workflow_config = get_perplexity_like_workflow_config()
    # yaml_file_name = "perplexity_like_workflow.yml"
    # workflow_config = get_perplexity_workflow_config()
    # yaml_file_name = "perplexity_workflow.yml"
    workflow_config = get_vertexai_workflow_config()
    yaml_file_name = "vertexai_workflow.yml"
    save_yaml_to_file(workflow_config.to_yaml(), yaml_file_name)
    user_workflow = UserWorkflow(
        user_id="sebastian.hsu@gmail.com",
        id="1234abcd",
        workflow_config_yaml=workflow_config.to_yaml(),
    )
    json_str = user_workflow.model_dump_json(indent=2)
    print(json_str)

    # 从文件读取 YAML
    loaded_yaml_str = load_yaml_from_file(yaml_file_name)
    question = "請問新北與高雄市的育兒津貼政策有哪些差異？"
    # question = "請問屏東的長照補助或津貼，與花蓮的差異為何？"
    # question = "我剛生了一對雙胞胎，政府有沒有補助？"
    # question = "中高齡失業，中央以及各縣市有哪些補助？"
    question = "如果寶寶已經滿兩歲了，由衛福部負責0至未滿2歲幼兒的育兒津貼，這個部分就不能再領了嗎？那我需要重新申請教育部負責2至未滿5歲幼兒的育兒津貼嗎？"
    question = "婚前健康檢查補助的費用及檢查項目內容有哪些？新北市跟雲林縣的差異是什麼？幫我做很詳實的比較。"
    # question = "我需要新竹縣所有鄉鎮的生育獎勵補助金額，請全部列出來"
    start_time = time.time()
    final_output = asyncio.run(
        run_workflow_from_yaml(
            loaded_yaml_str,
            question,
        )
    )
    end_time = time.time()
    print(f"Time taken: {end_time - start_time} seconds")
    # if final_output:
    #     print("Workflow completed. Final output:")
    #     print(final_output)
    # else:
    #     print("Workflow did not complete successfully.")



import asyncio
import json
from typing import List
import os
import aiohttp
from dotenv import load_dotenv
from google.oauth2 import service_account
from google.api_core.client_options import ClientOptions
from google.cloud import discoveryengine_v1 as discoveryengine
import litellm
from google.protobuf.json_format import MessageToDict, MessageToJson

from botrun_flow_lang.models.nodes.llm_node import get_api_key
from botrun_flow_lang.models.nodes.utils import (
    get_search_keywords,
    note_taking_scrape_results,
    scrape_single_url,
)

load_dotenv()

from typing import Any, AsyncGenerator, Dict, List
from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunCompletedEvent,
    NodeRunFailedEvent,
    NodeRunStreamEvent,
)
from botrun_flow_lang.models.variable import OutputVariable
from pydantic import Field


class VertexAiSearchNodeData(BaseNodeData):
    """
    @param results: 搜尋結果, 會是一個 List[Dict[str, Any]], dict 長這樣 {"title": "", "url": "", "snippet": ""}
    """

    type: NodeType = NodeType.VERTEX_AI_SEARCH
    search_query: str
    project_id: str
    location: str
    data_store_ids: List[str] = []
    output_variables: List[OutputVariable] = [OutputVariable(variable_name="results")]


class VertexAiSearchNode(BaseNode):
    data: VertexAiSearchNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        try:
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在生成搜尋關鍵字...\n",
                is_print=self.data.print_stream,
            )
            keywords = await get_search_keywords(
                self.replace_variables(self.data.search_query, variable_pool)
            )

            for keyword in keywords:
                yield NodeRunStreamEvent(
                    node_id=self.data.id,
                    node_title=self.data.title,
                    node_type=self.data.type.value,
                    chunk=f"- {keyword}\n",
                    is_print=self.data.print_stream,
                )

            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在搜尋關鍵字...\n",
                is_print=self.data.print_stream,
            )
            search_tasks = [self._search_keywords(keyword) for keyword in keywords]
            search_results = await asyncio.gather(*search_tasks)

            selected_urls = await self._choose_related_results(
                self.replace_variables(self.data.search_query, variable_pool),
                search_results,
            )
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在抓取網頁內容...\n",
                is_print=self.data.print_stream,
            )
            scrape_results = await scrape_vertexai_urls(selected_urls, search_results)
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在分析有效內容...\n",
                is_print=self.data.print_stream,
            )

            note_taking_results = await note_taking_scrape_results(
                self.replace_variables(self.data.search_query, variable_pool),
                scrape_results,
            )

            yield NodeRunCompletedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                outputs={
                    "results": note_taking_results,
                },
                is_print=self.data.print_complete,
            )

        except Exception as e:
            import traceback

            traceback.print_exc()
            yield NodeRunFailedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                error=str(e),
                is_print=True,
            )
            raise

    async def _choose_related_results(
        self, user_query: str, items: List[Dict[str, Any]]
    ) -> List[str]:
        """使用 LLM 選擇最相關的搜尋結果"""
        analyze_search_result_prompt = """請分析以下 Google 搜尋結果，參考 snippet 的內容，選出最相關的10個網頁連結，並且彼此內容的重覆性低。
請使用以下幾個優先權來選擇:
1. 如果有政府機關的網站，比如網址包含 gov.tw，請優先選擇政府機關的網站
2. 標題跟使用者問題最相關的
3. 標題包含使用者問題關鍵字的
4. 內容包含使用者問題關鍵字的
5. 網頁連結的內容重覆性低

使用者問題: {question}
搜尋結果: {items}

請務必只使用以下 JSON 格式嚴格回應，不要加上 markdown 格式:
{{
    "urls":["url1", "url2", "url3", "url4", "url5",..."url10"],
}}
"""

        model_name = "openai/gpt-4o-2024-08-06"
        response = litellm.completion(
            model=model_name,
            response_format={"type": "json_object"},
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant designed to output JSON.",
                },
                {
                    "role": "user",
                    "content": analyze_search_result_prompt.format(
                        question=user_query,
                        items=json.dumps(items, ensure_ascii=False, indent=2),
                    ),
                },
            ],
            api_key=get_api_key(model_name),
        )

        try:
            selected_urls = json.loads(response.choices[0].message.content)["urls"]
            print(f"selected_urls: {selected_urls}")
            return selected_urls
        except Exception as e:
            print(f"Error parsing LLM response for URL selection: {str(e)}")
            return []

    async def _search_keywords(self, keyword: str) -> List[Dict[str, Any]]:
        """對單個關鍵字進行搜尋"""
        try:
            vertex_search = VertexAISearch()
            all_results = []

            # 對每個 data store 進行平行搜尋
            async def search_data_store(data_store_id: str) -> List[Dict[str, Any]]:
                try:
                    results = vertex_search.vertex_search(
                        project_id=self.data.project_id,
                        location=self.data.location,
                        data_store_id=data_store_id,
                        search_query=keyword,
                    )
                    return results.get("results", [])
                except Exception as e:
                    print(f"Error searching {data_store_id}: {str(e)}")
                    return []

            # 創建所有 data store 的搜尋任務
            search_tasks = [
                search_data_store(data_store_id)
                for data_store_id in self.data.data_store_ids
            ]

            # 平行執行所有搜尋任務
            results_list = await asyncio.gather(*search_tasks)

            # 合併所有結果
            for results in results_list:
                if results:
                    all_results.extend(results)

            return all_results

        except Exception as e:
            print(f"Error in _search_keywords: {str(e)}")
            return []


async def scrape_vertexai_urls(
    selected_urls: List[str], vertexai_search_results: List[Dict]
) -> List[Dict[str, Any]]:
    """並行抓取所有 URL 的內容"""
    scrape_candidates = []
    for keyword_search_result in vertexai_search_results:
        for result in keyword_search_result:
            if result["url"] in selected_urls:
                scrape_candidates.append(result)
    scrape_tasks = []
    for scrape_candidate in scrape_candidates:
        scrape_tasks.append(scrape_single_url(scrape_candidate["url"]))

    # 一次性創建所有 URL 的抓取任務
    scrape_tasks = [scrape_single_url(url) for url in selected_urls]

    # 同時執行所有抓取任務
    scrape_results = await asyncio.gather(*scrape_tasks)
    scrape_results = [
        scrape_result
        for scrape_result in scrape_results
        if scrape_result["status"] == "success"
    ]

    # 轉換為原來的輸出格式
    return scrape_results


class VertexAISearch:
    def __init__(self):
        # 使用與 FirestoreBase 相同的認證方式
        google_service_account_key_path = os.getenv(
            "GOOGLE_APPLICATION_CREDENTIALS_FOR_FASTAPI",
            "/app/keys/scoop-386004-d22d99a7afd9.json",
        )
        self.credentials = service_account.Credentials.from_service_account_file(
            google_service_account_key_path,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )

    def vertex_search(
        self,
        project_id: str,
        location: str,
        data_store_id: str,
        search_query: str,
    ) -> List[discoveryengine.ConverseConversationResponse]:
        try:
            client_options = (
                ClientOptions(api_endpoint=f"{location}-discoveryengine.googleapis.com")
                if location != "global"
                else None
            )

            client = discoveryengine.SearchServiceClient(
                credentials=self.credentials, client_options=client_options
            )

            serving_config = client.serving_config_path(
                project=project_id,
                location=location,
                data_store=data_store_id,
                serving_config="default_config",
            )

            # 添加 filter 來排除特定檔案類型
            file_filter = 'NOT(mimeType:ANY("application/pdf"))'
            # 建立搜尋請求
            request = discoveryengine.SearchRequest(
                serving_config=serving_config,
                query=search_query,
                # filter=file_filter,
                page_size=10,
            )

            response = client.search(request)

            # 或者使用 JSON 格式（更易讀）
            # print("Response in JSON:")
            # print(MessageToJson(response._pb, indent=2))

            results = []
            for result in response.results:
                # 打印單個結果的詳細信息
                # print("\nSingle Result Details:")
                # print(MessageToJson(result._pb, indent=2))

                document_data = result.document.derived_struct_data
                results.append(
                    {
                        "title": document_data.get("title", "No title"),
                        "url": document_data.get("link", "No link"),
                        "snippet": document_data.get("snippets", "No snippet")[0][
                            "snippet"
                        ],
                        "fileFormat": document_data.get("fileFormat", ""),
                    }
                )

            return {"results": results}

        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error occurred: {str(e)}")
            return {"results": []}

    def get_document_content(
        self, project_id: str, location: str, data_store_id: str, document_id: str
    ):
        client = discoveryengine.DocumentServiceClient(credentials=self.credentials)

        name = client.document_path(
            project=project_id,
            location=location,
            data_store=data_store_id,
            branch="default_branch",
            document=document_id,
        )

        try:
            document = client.get_document(name=name)
            return document
        except Exception as e:
            print(f"Error getting document: {e}")
            return None


def main():
    # 從環境變數讀取設定
    project_id = "scoop-386004"
    location = "global"
    data_store_ids = [
        "tw-gov-welfare_1730944342934",
        # "tw-gov-welfare-files_1730960173279",
    ]

    # 測試用的搜尋查詢
    search_queries = ["婚前健康檢查", "相關新聞"]
    search_queries = ["苗栗", "生育補助", "相關新聞"]

    # if not all([project_id, data_store_id]):
    #     print("Error: Missing required environment variables.")
    #     print("Please ensure GCP_PROJECT_ID and VERTEX_SEARCH_DATASTORE are set.")
    #     return

    for data_store_id in data_store_ids:
        try:
            vertex_search = VertexAISearch()
            results = vertex_search.vertex_search(
                project_id=project_id,
                location=location,
                data_store_id=data_store_id,
                search_query=" ".join(search_queries),
            )
            print(results)
        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error occurred: {str(e)}")


if __name__ == "__main__":
    main()



from typing import Any, AsyncGenerator, Dict, List
from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunCompletedEvent,
    NodeRunFailedEvent,
    NodeRunStreamEvent,
)
import litellm
import json
import aiohttp
import asyncio
from botrun_flow_lang.models.nodes.llm_node import get_api_key
from urllib.parse import quote

from botrun_flow_lang.models.nodes.utils import (
    get_search_keywords,
    note_taking_scrape_results,
    scrape_urls,
)
from botrun_flow_lang.models.variable import OutputVariable
from pydantic import Field
import time


class SearchAndScrapeNodeData(BaseNodeData):
    """
    @param results: 筆記的結果, 會是一個 List[Dict[str, Any]], dict 長這樣{"question": "","url": "","title": "","note": ""}
    """

    type: NodeType = NodeType.SEARCH_AND_SCRAPE
    search_query: str
    output_variables: List[OutputVariable] = [OutputVariable(variable_name="results")]


class SearchAndScrapeNode(BaseNode):
    data: SearchAndScrapeNodeData

    async def _search_question(
        self, question: str, session: aiohttp.ClientSession
    ) -> Dict[str, Any]:
        """執行單個問題的搜尋"""
        url = "https://botrun-flow-lang-fastapi-prod-36186877499.asia-east1.run.app/api/search"
        try:
            async with session.post(
                url,
                json={"query": question, "num": 10},
                headers={"Content-Type": "application/json"},
            ) as response:
                if response.status == 200:
                    result = await response.json()
                    return {
                        "question": question,
                        "status": "success",
                        "items": result["items"],
                    }
                else:
                    return {
                        "question": question,
                        "status": "error",
                        "error": f"Search failed with status {response.status}",
                    }
        except Exception as e:
            return {"question": question, "status": "error", "error": str(e)}

    async def _choose_related_results(
        self, user_query: str, search_results: List[Dict[str, Any]]
    ) -> List[str]:
        """使用 LLM 選擇最相關的搜尋結果"""
        analyze_search_result_prompt = """請分析以下 Google 搜尋結果，參考 snippet 的內容，選出最相關的10個網頁連結，並且彼此內容的重覆性低。
請使用以下幾個優先權來選擇:
1. 如果有政府機關的網站，比如網址包含 gov.tw，請優先選擇政府機關的網站
2. 標題跟使用者問題最相關的
3. 標題包含使用者問題關鍵字的
4. 內容包含使用者問題關鍵字的
5. 網頁連結的內容重覆性低

使用者問題: {question}
搜尋結果: {items}

請務必只使用以下 JSON 格式嚴格回應，不要加上 markdown 格式:
{{
    "urls":["url1", "url2", "url3", "url4", "url5",..."url10"],
}}
"""
        items = []
        for result in search_results:
            if result["status"] == "success":
                items.extend(result["items"])

        model_name = "openai/gpt-4o-2024-08-06"
        response = litellm.completion(
            model=model_name,
            response_format={"type": "json_object"},
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant designed to output JSON.",
                },
                {
                    "role": "user",
                    "content": analyze_search_result_prompt.format(
                        question=user_query,
                        items=json.dumps(items, ensure_ascii=False, indent=2),
                    ),
                },
            ],
            api_key=get_api_key(model_name),
        )

        try:
            selected_urls = json.loads(response.choices[0].message.content)["urls"]
            print(f"selected_urls: {selected_urls}")
            return selected_urls
        except Exception as e:
            print(f"Error parsing LLM response for URL selection: {str(e)}")
            return []

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        try:
            time_message = []
            time_1 = time.time()
            # 1. 生成搜尋問題
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在生成搜尋關鍵字...\n",
                is_print=self.data.print_stream,
            )
            keywords = await get_search_keywords(
                self.replace_variables(self.data.search_query, variable_pool)
            )
            time_2 = time.time()
            time_message.append(f"生成搜尋問題: {time_2 - time_1:.2f} 秒")
            for keyword in keywords:
                yield NodeRunStreamEvent(
                    node_id=self.data.id,
                    node_title=self.data.title,
                    node_type=self.data.type.value,
                    chunk=f"- {keyword}\n",
                    is_print=self.data.print_stream,
                )

            # 2. 執行搜尋
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在將問題交由 Google 搜尋...\n",
                is_print=self.data.print_stream,
            )
            async with aiohttp.ClientSession() as session:
                search_tasks = [
                    self._search_question(question, session) for question in keywords
                ]
                search_results = await asyncio.gather(*search_tasks)
            time_3 = time.time()
            time_message.append(f"將問題交由 Google 搜尋: {time_3 - time_2:.2f} 秒")
            # 使用 asyncio.gather 同步執行所有分析任務
            selected_urls = await self._choose_related_results(
                self.replace_variables(self.data.search_query, variable_pool),
                search_results,
            )
            time_4 = time.time()
            time_message.append(f"選擇相關搜尋結果: {time_4 - time_3:.2f} 秒")
            # print(analyzed_search_results)

            # 3. 抓取網頁內容
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在抓取網頁內容...\n",
                is_print=self.data.print_stream,
            )
            scrape_results = await scrape_urls(selected_urls)
            time_5 = time.time()
            time_message.append(f"抓取網頁內容: {time_5 - time_4:.2f} 秒")
            # 4. 進行筆記
            yield NodeRunStreamEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                chunk=f"\n正在分析有效內容...\n",
                is_print=self.data.print_stream,
            )

            note_taking_results = await note_taking_scrape_results(
                self.replace_variables(self.data.search_query, variable_pool),
                scrape_results,
            )
            time_6 = time.time()
            time_message.append(f"進行筆記: {time_6 - time_5:.2f} 秒")
            # for result in note_taking_results:
            #     yield NodeRunStreamEvent(
            #         node_id=self.data.id,
            #         node_title=self.data.title,
            #         node_type=self.data.type.value,
            #         chunk=f"完成筆記: {result['question']} ({len(result['note_taking_results'])} 個筆記)\n",
            #         is_print=self.data.print_stream,
            #     )
            for message in time_message:
                print(message)
            # 5. 返回最終結果
            yield NodeRunCompletedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                outputs={
                    "results": note_taking_results,
                },
                is_print=self.data.print_complete,
            )

        except Exception as e:
            yield NodeRunFailedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                error=str(e),
                is_print=True,
            )
            raise



import ast
from typing import Dict, Any, AsyncGenerator
from pydantic import Field, field_validator

from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunCompletedEvent,
)


class CodeNodeData(BaseNodeData):
    """
    function回傳一個 dict，key 是 output_variables 的 variable_name，value 是對應的值
    """

    type: NodeType = NodeType.CODE
    code: str = Field(..., description="Python3 code to be executed")
    input_variables: list[InputVariable] = Field(default_factory=list)
    output_variables: list[OutputVariable] = Field(default_factory=list)

    @field_validator("code")
    def validate_code(cls, v):
        try:
            ast.parse(v)
        except SyntaxError as e:
            raise ValueError(f"Invalid Python code: {str(e)}")
        return v


class CodeNode(BaseNode):
    data: CodeNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        input_args = {}
        for input_var in self.data.input_variables:
            input_args[input_var.variable_name] = self.get_variable(
                variable_pool, input_var.node_id, input_var.variable_name
            )

        # Execute the code
        try:
            exec_globals = {}
            exec(self.data.code, exec_globals)
            main_func = exec_globals.get("main")
            if not main_func:
                raise ValueError("No 'main' function defined in the code")

            result = main_func(**input_args)

            if not isinstance(result, dict):
                raise ValueError("The 'main' function must return a dictionary")

            # Prepare output
            output = {}
            for output_var in self.data.output_variables:
                if output_var.variable_name in result:
                    output[output_var.variable_name] = result[output_var.variable_name]
                else:
                    raise ValueError(
                        f"Output variable '{output_var.variable_name}' not found in the result"
                    )

            yield NodeRunCompletedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                outputs=output,
                complete_output=self.data.complete_output,
                is_print=self.data.print_complete,
            )
        except Exception as e:
            raise RuntimeError(f"Error executing {self.data.title} code: {str(e)}")



from typing import List
from pydantic import BaseModel, Field


class Hatch(BaseModel):
    """
    @user_prompt_prefix: 每次的 user prompt 前面都會加入這段文字
    @search_domain_filter: 搜尋的網域限制, 目前只有針對 perplexit 有效, 範例：["*.gov.tw", "-*.gov.cn"]
    """

    user_id: str
    id: str
    prompt_template: str
    user_prompt_prefix: str = ""
    name: str = ""  # 将 name 设为可选字段，默认为空字符串
    is_default: bool = False
    enable_search: bool = False
    related_question_prompt: str = ""
    search_vendor: str = "perplexity"
    search_domain_filter: List[str] = []



from pydantic import field_validator
from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from typing import Dict, Any, List, AsyncGenerator

from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.nodes.event import NodeEvent, NodeRunCompletedEvent


class StartNodeData(BaseNodeData):
    type: NodeType = NodeType.START
    input_variables: List[InputVariable] = []
    output_variables: List[OutputVariable] = [
        OutputVariable(variable_name="user_input"),
        OutputVariable(variable_name="history"),
    ]

    @field_validator("input_variables")
    def validate_input_variables(cls, v):
        assert len(v) == 0, "StartNode must have 0 input variables"
        return v

    @field_validator("output_variables")
    def validate_output_variables(cls, v):
        assert len(v) == 2, "StartNode must have exactly 2 output variable"
        assert (
            v[0].variable_name == "user_input"
        ), "StartNode's output variable must be named 'user_input'"
        assert (
            v[1].variable_name == "history"
        ), "StartNode's output variable must be named 'history'"
        return v


class StartNode(BaseNode):
    data: StartNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        yield NodeRunCompletedEvent(
            node_id=self.data.id,
            node_title=self.data.title,
            node_type=self.data.type.value,
            outputs={
                "user_input": variable_pool.get(self.data.id, {}).get("user_input", ""),
                "history": variable_pool.get(self.data.id, {}).get("history", []),
            },
            complete_output=self.data.complete_output,
            is_print=self.data.print_start,
        )



from enum import Enum
from typing import Dict, Any, AsyncGenerator, Optional, List
from pydantic import BaseModel, Field, field_validator
import aiohttp
import json

from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunCompletedEvent,
)


class AuthorizationType(str, Enum):
    NONE = "none"
    API_KEY = "api-key"


class ApiKeyType(str, Enum):
    BEARER = "bearer"
    BASIC = "basic"
    CUSTOM = "custom"


class BodyType(str, Enum):
    NONE = "none"
    JSON = "json"
    FORM_DATA = "form-data"
    X_WWW_FORM_URLENCODED = "x-www-form-urlencoded"
    RAW = "raw"
    BINARY = "binary"


class HttpMethod(str, Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"


class AuthorizationConfig(BaseModel):
    api_key: str
    type: ApiKeyType


class Authorization(BaseModel):
    type: AuthorizationType = AuthorizationType.NONE
    config: Optional[AuthorizationConfig] = None


class Body(BaseModel):
    type: BodyType = BodyType.NONE
    data: Any = None


class Timeout(BaseModel):
    total: float = 30.0


class HttpRequestNodeData(BaseNodeData):
    type: NodeType = NodeType.HTTP_REQUEST
    authorization: Authorization = Field(default_factory=Authorization)
    body: Body = Field(default_factory=Body)
    headers: str = ""
    method: HttpMethod = HttpMethod.GET
    params: str = ""
    timeout: Timeout = Field(default_factory=Timeout)
    url: str
    input_variables: List[InputVariable] = Field(default_factory=list)
    output_variables: List[OutputVariable] = Field(
        default_factory=lambda: [
            OutputVariable(variable_name="status_code"),
            OutputVariable(variable_name="body"),
            OutputVariable(variable_name="headers"),
        ]
    )

    @field_validator("output_variables")
    def validate_output_variables(cls, v):
        required_vars = {"status_code", "body", "headers"}
        actual_vars = {ov.variable_name for ov in v}
        assert required_vars.issubset(
            actual_vars
        ), f"HttpRequestNode must have output variables: {required_vars}"
        return v


class HttpRequestNode(BaseNode):
    data: HttpRequestNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        try:
            headers = self._parse_headers(variable_pool)
            params = self._parse_params(variable_pool)
            body = self._prepare_body(variable_pool)

            url = self.replace_variables(self.data.url, variable_pool)
            print(f"[HttpRequestNode]url: {url}")
            print(f"[HttpRequestNode]params: {params}")

            timeout = aiohttp.ClientTimeout(total=self.data.timeout.total)

            async with aiohttp.ClientSession(timeout=timeout) as session:
                async with session.request(
                    method=self.data.method.value,
                    url=url,
                    headers=headers,
                    params=params,
                    data=body,
                ) as response:
                    status_code = response.status
                    response_body = await response.text()
                    response_headers = dict(response.headers)

            yield NodeRunCompletedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                outputs={
                    "status_code": status_code,
                    "body": response_body,
                    "headers": response_headers,
                },
                complete_output=self.data.complete_output,
                is_print=self.data.print_complete,
            )
        except Exception as e:
            print(f"Error executing HTTP request: url: {url}, params: {params}")
            print(f"http request exception: {str(e)}")
            # print(f"params: {params}")
            # import traceback

            # traceback.print_exc()
            yield NodeRunCompletedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                outputs={
                    "status_code": "",
                    "body": f"Error executing HTTP request: {str(e)}",
                    "headers": {},
                },
                complete_output=self.data.complete_output,
                is_print=self.data.print_complete,
            )

            # raise RuntimeError(f"Error executing HTTP request: {str(e)}")

    def _parse_headers(self, variable_pool: Dict[str, Dict[str, Any]]) -> dict:
        headers = {}
        if self.data.headers:
            for line in self.data.headers.split("\n"):
                if ":" in line:
                    key, value = line.split(":", 1)
                    headers[key.strip()] = self.replace_variables(
                        value.strip(), variable_pool
                    )

        if self.data.authorization.type == AuthorizationType.API_KEY:
            if self.data.authorization.config:
                api_key = self.replace_variables(
                    self.data.authorization.config.api_key, variable_pool
                )
                if self.data.authorization.config.type == ApiKeyType.BEARER:
                    headers["Authorization"] = f"Bearer {api_key}"
                elif self.data.authorization.config.type == ApiKeyType.BASIC:
                    headers["Authorization"] = f"Basic {api_key}"
                elif self.data.authorization.config.type == ApiKeyType.CUSTOM:
                    headers["Authorization"] = api_key
        if self.data.body.type == BodyType.JSON:
            headers["Content-Type"] = "application/json"
        return headers

    def _parse_params(self, variable_pool: Dict[str, Dict[str, Any]]) -> dict:
        params = {}
        if self.data.params:
            for line in self.data.params.split("\n"):
                if ":" in line:
                    key, value = line.split(":", 1)
                    params[key.strip()] = self.replace_variables(
                        value.strip(), variable_pool
                    )
        return params

    def _prepare_body(self, variable_pool: Dict[str, Dict[str, Any]]) -> Any:
        if self.data.body.type == BodyType.NONE:
            return None
        elif self.data.body.type == BodyType.JSON:
            new_data = {}
            for key, value in self.data.body.data.items():
                if isinstance(value, str):
                    new_data[key] = self.replace_variables(value, variable_pool)
                else:
                    new_data[key] = value
            return json.dumps(new_data)  # 將字典轉換為 JSON 字符串
        elif self.data.body.type == BodyType.FORM_DATA:
            return aiohttp.FormData(
                {
                    k: self.replace_variables(v, variable_pool)
                    for k, v in self.data.body.data.items()
                }
            )
        elif self.data.body.type == BodyType.X_WWW_FORM_URLENCODED:
            return aiohttp.FormData(
                {
                    k: self.replace_variables(v, variable_pool)
                    for k, v in self.data.body.data.items()
                },
                quote_fields=False,
            )
        elif self.data.body.type == BodyType.RAW:
            return self.replace_variables(self.data.body.data, variable_pool)
        elif self.data.body.type == BodyType.BINARY:
            # Implement binary data handling if needed
            raise NotImplementedError("Binary body type is not implemented")



from .code_node import CodeNode, CodeNodeData
from .start_node import StartNode, StartNodeData
from .llm_node import LLMNode, LLMNodeData
from .answer_node import AnswerNode, AnswerNodeData
from .end_node import EndNode, EndNodeData
from .http_request_node import HttpRequestNode, HttpRequestNodeData
from .iteration_node import IterationNode, IterationNodeData



from typing import Optional, Dict, Any
from pydantic import BaseModel, Field
import uuid
from botrun_flow_lang.models.workflow_config import WorkflowConfig


class UserWorkflow(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    user_id: str
    workflow_config_yaml: str



import os

from botrun_flow_lang.services.user_workflow.user_workflow_fs_store import (
    UserWorkflowFsStore,
)


def user_workflow_store_factory():
    env_name = os.getenv("ENV_NAME", "dev")
    return UserWorkflowFsStore(env_name)



import unittest
import asyncio
from botrun_flow_lang.models.nodes.base_node import NodeType
from botrun_flow_lang.models.nodes.start_node import StartNode, StartNodeData
from botrun_flow_lang.models.nodes.event import NodeRunCompletedEvent
import uuid


class TestStartNode(unittest.TestCase):
    def test_start_node_data_creation(self):
        start_node = StartNodeData(title="Test Start")

        self.assertEqual(start_node.type, NodeType.START)
        self.assertEqual(start_node.title, "Test Start")

        # Test that id is a string and can be parsed as a valid UUID
        self.assertIsInstance(start_node.id, str)
        self.assertTrue(uuid.UUID(start_node.id, version=4))

    async def async_test_start_node_run(self):
        start_node = StartNode(data=StartNodeData(title="Test Start"))
        variable_pool = {"user_input": "Hello, world!"}

        async for event in start_node.run(variable_pool):
            self.assertIsInstance(event, NodeRunCompletedEvent)
            self.assertEqual(event.outputs, {"user_input": "Hello, world!"})

    def test_start_node_run(self):
        asyncio.run(self.async_test_start_node_run())


if __name__ == "__main__":
    unittest.main()



from fastapi import APIRouter, HTTPException, Depends, Query, Body
from botrun_flow_lang.api.user_setting_api import get_user_setting_store
from botrun_flow_lang.services.hatch.hatch import Hatch
from botrun_flow_lang.services.hatch.hatch_factory import hatch_store_factory
from botrun_flow_lang.services.hatch.hatch_fs_store import HatchFsStore
from typing import List

from botrun_flow_lang.services.user_setting.user_setting_fs_store import (
    UserSettingFsStore,
)

router = APIRouter()


async def get_hatch_store():
    return hatch_store_factory()


@router.post("/hatch", response_model=Hatch)
async def create_hatch(hatch: Hatch, store: HatchFsStore = Depends(get_hatch_store)):
    success, created_hatch = await store.set_hatch(hatch)
    if not success:
        raise HTTPException(status_code=500, detail="Failed to create hatch")
    return created_hatch


@router.put("/hatch/{hatch_id}", response_model=Hatch)
async def update_hatch(
    hatch_id: str, hatch: Hatch, store: HatchFsStore = Depends(get_hatch_store)
):
    existing_hatch = await store.get_hatch(hatch_id)
    if not existing_hatch:
        raise HTTPException(status_code=404, detail="Hatch not found")
    hatch.id = hatch_id
    success, updated_hatch = await store.set_hatch(hatch)
    if not success:
        raise HTTPException(status_code=500, detail="Failed to update hatch")
    return updated_hatch


@router.delete("/hatch/{hatch_id}")
async def delete_hatch(hatch_id: str, store: HatchFsStore = Depends(get_hatch_store)):
    success = await store.delete_hatch(hatch_id)
    if not success:
        raise HTTPException(
            status_code=500,
            detail={"success": False, "message": f"Failed to delete hatch {hatch_id}"},
        )
    return {"success": True, "message": f"Hatch {hatch_id} deleted successfully"}


@router.get("/hatch/{hatch_id}", response_model=Hatch)
async def get_hatch(hatch_id: str, store: HatchFsStore = Depends(get_hatch_store)):
    hatch = await store.get_hatch(hatch_id)
    if not hatch:
        raise HTTPException(status_code=404, detail="Hatch not found")
    return hatch


@router.get("/hatches", response_model=List[Hatch])
async def get_hatches(
    user_id: str,
    offset: int = Query(0, ge=0),
    limit: int = Query(20, ge=1, le=100),
    hatch_store=Depends(get_hatch_store),
):
    hatches, error = await hatch_store.get_hatches(user_id, offset, limit)
    if error:
        raise HTTPException(status_code=500, detail=error)
    return hatches


@router.get("/hatch/default/{user_id}", response_model=Hatch)
async def get_default_hatch(
    user_id: str, store: HatchFsStore = Depends(get_hatch_store)
):
    default_hatch = await store.get_default_hatch(user_id)
    if not default_hatch:
        raise HTTPException(status_code=404, detail="Default hatch not found")
    return default_hatch


@router.post("/hatch/set_default")
async def set_default_hatch(
    user_id: str = Body(...),
    hatch_id: str = Body(...),
    store: HatchFsStore = Depends(get_hatch_store),
):
    success, message = await store.set_default_hatch(user_id, hatch_id)
    if not success:
        raise HTTPException(status_code=500, detail=message)
    return {"success": True, "message": message}


@router.get("/hatches/statistics")
async def get_hatches_statistics(
    user_setting_store: UserSettingFsStore = Depends(get_user_setting_store),
    hatch_store: HatchFsStore = Depends(get_hatch_store),
):
    """Get statistics about hatches across all users.

    Returns:
        dict: Contains total hatch count and per-user hatch counts
    """
    try:
        # Get all user IDs
        user_ids = await user_setting_store.get_all_user_ids()

        # Initialize statistics
        all_hatches = []
        total_count = 0

        # Get hatch counts for each user
        for user_id in user_ids:
            hatches, _ = await hatch_store.get_hatches(user_id)
            count = len(hatches)
            if count > 0:  # Only include users who have hatches
                all_hatches.append({"user_id": user_id, "hatches_count": count})
                total_count += count

        return {"all_hatches_count": total_count, "all_hatches": all_hatches}
    except Exception as e:
        raise HTTPException(
            status_code=500, detail=f"Failed to fetch hatch statistics: {str(e)}"
        )



import asyncio
import json
from typing import Any, Dict, List
from urllib.parse import quote

import aiohttp
import litellm
from yarl import URL

from botrun_flow_lang.models.nodes.llm_node import get_api_key


async def get_search_keywords(search_query: str) -> List[str]:
    """使用 LLM 生成搜尋關鍵字"""
    generate_questions_prompt = """
    你是一個專業的調查員，你會依據以下問題，去網路上搜尋相關資料，並且回答使用者。
    當使用者輸入一個問題時，你會
    1. 理解查詢：理解用戶輸入的查詢。這不僅僅是簡單的關鍵字匹配，而是深入分析查詢的上下文和意圖，以便更準確地理解用戶需求。
    2. 構建查詢：在理解查詢後，你會重構查詢以應其搜索和分析模型。這包括將用戶的自然語言問題轉換為可以在網路上有效搜索的訊息格式，從而提高搜索效率和結果的相關性。
    3. 條列重構查詢：將重構後的查詢，條列成3組搜尋此問題的關鍵字，同一組可以有多個關鍵字，每組關鍵字之間用 空格 隔開。
    4. 每組關鍵字最後面都會加上"相關新聞"。

    以下是使用者輸入的問題:
    {search_query}

    請使用以下 JSON 格式嚴格回應,只包含問題內容,不要使用 markdown 的語法:
    {{
        "keywords":[
            "第1組關鍵字",
            "第2組關鍵字",
            ...
            "最後一組關鍵字"
        ]
    }}
""".format(
        search_query=search_query
    )

    model_name = "anthropic/claude-3-5-sonnet-20241022"
    response = litellm.completion(
        model=model_name,
        messages=[{"role": "user", "content": generate_questions_prompt}],
        api_key=get_api_key(model_name),
    )
    return json.loads(response.choices[0].message.content)["keywords"]


async def scrape_urls(selected_urls: List[str]) -> List[Dict[str, Any]]:
    """並行抓取所有 URL 的內容"""
    # 一次性創建所有 URL 的抓取任務
    scrape_tasks = [scrape_single_url(url) for url in selected_urls]

    # 同時執行所有抓取任務
    scrape_results = await asyncio.gather(*scrape_tasks)
    scrape_results = [
        scrape_result
        for scrape_result in scrape_results
        if scrape_result["status"] == "success"
    ]

    # 轉換為原來的輸出格式
    return scrape_results


async def scrape_single_url(url: str) -> Dict[str, Any]:
    """抓取單個 URL 的內容"""
    try:
        if "%" not in url:
            quoted_url = quote(url, safe="")
        else:
            quoted_url = url
        scrape_url = f"https://botrun-crawler-fastapi-prod-36186877499.asia-east1.run.app/scrape?url={quoted_url}"
        scrape_url = URL(scrape_url, encoded=True)
        async with aiohttp.ClientSession() as session:
            async with session.get(scrape_url) as response:
                if response.status == 200:
                    body = await response.json()
                    print(f"[scrape_single_url] url: {url}")
                    print(
                        f"[scrape_single_url] content: {body['data']['markdown'][:100]}"
                    )
                    return {
                        "url": url,
                        "title": body["data"]["metadata"]["title"],
                        "content": body["data"]["markdown"],
                        "status": "success",
                    }
                else:
                    return {
                        "url": url,
                        "status": "error",
                        "error": f"Scraping failed with status {response.status}",
                    }
    except Exception as e:
        return {"url": url, "status": "error", "error": str(e)}


async def note_taking_single_result(
    user_query: str, scrape_result: Dict[str, Any]
) -> Dict[str, Any]:
    """對單個抓取結果進行筆記"""
    note_taker_prompt = """你是一位資料記錄者，並具有基礙程式能力，瞭解 HTML 語法，請分析以下網頁內容，做出詳實記錄。

網頁URL: {url}
網頁標題: {title}
網頁內容: {markdown}

請：
1. 去除不必要的 HTML 資料，
2. 去除與使用者問題無關的行銷內容
3. 去除廣告的內容
4. 去除看起來像是header, footer, sidebar的內容
5. 去除看起來像是版權宣告的內容
6. 去除看起來像是目錄的內容
7. 去除看起來像是導覽列的內容
8. 去除首起來像是連到其它文章的內容

你是記錄者，所以你不要加上任何自己的主觀意見，只做完上述工作後，留下詳實的記錄內容。

請使用以下 JSON 格式嚴格回應，不要附加任何其它文字，不要加上 markdown 的語法:
{{
"url": "網頁URL",
"title": "網頁標題",
"note": "詳實的記錄內容"
}}
"""

    model_name = "gemini/gemini-1.5-flash"
    try:
        response = litellm.completion(
            model=model_name,
            response_format={"type": "json_object"},
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant designed to output JSON.",
                },
                {
                    "role": "user",
                    "content": note_taker_prompt.format(
                        # user_query=user_query,
                        url=scrape_result["url"],
                        title=scrape_result["title"],
                        markdown=scrape_result["content"],
                    ),
                },
            ],
            api_key=get_api_key(model_name),
        )

        result = json.loads(response.choices[0].message.content)
        print(f"[note_taking_single_result] url: {result['url']}")
        print(f"[note_taking_single_result] title: {result['title']}")
        print(f"[note_taking_single_result] note: {result['note']}")
        return result if result.get("note") else None

    except Exception as e:
        print(f"Error in note taking for URL {scrape_result['url']}: {str(e)}")
        return None


async def note_taking_scrape_results(
    user_query: str, scrape_results: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    """並行處理所有抓取結果的筆記"""
    # 收集所有需要做筆記的任務
    note_tasks = []

    for scrape_result in scrape_results:
        note_tasks.append(note_taking_single_result(user_query, scrape_result))

    # 一次性執行所有筆記任務
    notes = await asyncio.gather(*note_tasks)
    return [note for note in notes if note is not None and note["url"] != ""]
    # 組織結果



from typing import Dict, Any, List, Union
import uuid
from pydantic import BaseModel, field_validator, Field
import yaml
from botrun_flow_lang.models.botrun_app import BotrunApp, BotrunAppMode
from botrun_flow_lang.models.nodes.code_node import CodeNodeData
from botrun_flow_lang.models.nodes.http_request_node import HttpRequestNodeData
from botrun_flow_lang.models.nodes.iteration_node import IterationNodeData
from botrun_flow_lang.models.nodes.perplexity_node import PerplexityNodeData
from botrun_flow_lang.models.nodes.vertex_ai_search_node import VertexAiSearchNodeData
from botrun_flow_lang.models.workflow import WorkflowData, Workflow, WorkflowItem
from botrun_flow_lang.models.nodes.start_node import StartNodeData
from botrun_flow_lang.models.nodes.llm_node import LLMNodeData
from botrun_flow_lang.models.nodes.end_node import EndNodeData
from botrun_flow_lang.models.nodes.answer_node import AnswerNodeData
from botrun_flow_lang.models.nodes.search_and_scrape_node import SearchAndScrapeNodeData
from botrun_flow_lang.models.nodes.base_node import NodeType, BaseNodeData

NodeData = Union[
    StartNodeData,
    LLMNodeData,
    EndNodeData,
    AnswerNodeData,
    CodeNodeData,
    HttpRequestNodeData,
    IterationNodeData,
    SearchAndScrapeNodeData,
    PerplexityNodeData,
]


class WorkflowConfig(BaseModel):
    botrun_app: BotrunApp
    workflow: WorkflowData

    @field_validator("workflow")
    def validate_workflow(cls, v, values):
        if "botrun_app" not in values.data:
            raise ValueError("botrun_app must be provided")

        botrun_app = values.data["botrun_app"]
        nodes = v.nodes

        if not nodes:
            raise ValueError("Workflow must have at least one node")

        if botrun_app.mode == BotrunAppMode.WORKFLOW:
            if not cls._ends_with_node_type(nodes, EndNodeData):
                raise ValueError("Workflow mode must end with an EndNode")
        elif botrun_app.mode == BotrunAppMode.CHATBOT:
            if not cls._ends_with_node_type(nodes, AnswerNodeData):
                raise ValueError("Chatbot mode must end with an AnswerNode")

        return v

    @classmethod
    def _ends_with_node_type(cls, nodes, node_type):
        if isinstance(nodes[-1], list):
            return cls._ends_with_node_type(nodes[-1], node_type)
        return isinstance(nodes[-1], node_type)

    def to_yaml(self) -> str:
        return yaml.dump(
            self.model_dump(mode="json"),
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=False,
        )

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "WorkflowConfig":
        data = yaml.safe_load(yaml_str)
        botrun_app = BotrunApp(**data["botrun_app"])

        nodes = cls._parse_nodes(data["workflow"]["nodes"])
        workflow = WorkflowData(nodes=nodes)
        return cls(botrun_app=botrun_app, workflow=workflow)

    @staticmethod
    def _parse_nodes(
        node_data_list: List[Union[Dict, List]]
    ) -> List[Union[NodeData, List]]:
        nodes = []
        for item in node_data_list:
            if isinstance(item, list):
                nodes.append(
                    [WorkflowConfig._create_node(item[0])]
                    + WorkflowConfig._parse_nodes(item[1:])
                )
            else:
                nodes.append(WorkflowConfig._create_node(item))
        return nodes

    @staticmethod
    def _create_node(node_data: Dict[str, Any]) -> NodeData:
        node_type = NodeType(node_data["type"])
        if node_type == NodeType.START:
            return StartNodeData(**node_data)
        elif node_type == NodeType.LLM:
            return LLMNodeData(**node_data)
        elif node_type == NodeType.END:
            return EndNodeData(**node_data)
        elif node_type == NodeType.ANSWER:
            return AnswerNodeData(**node_data)
        elif node_type == NodeType.CODE:
            return CodeNodeData(**node_data)
        elif node_type == NodeType.HTTP_REQUEST:
            return HttpRequestNodeData(**node_data)
        elif node_type == NodeType.ITERATION:
            return IterationNodeData(**node_data)
        elif node_type == NodeType.SEARCH_AND_SCRAPE:
            return SearchAndScrapeNodeData(**node_data)
        elif node_type == NodeType.VERTEX_AI_SEARCH:
            return VertexAiSearchNodeData(**node_data)
        elif node_type == NodeType.PERPLEXITY:
            return PerplexityNodeData(**node_data)
        else:
            raise ValueError(f"Unknown node type: {node_type}")

    def to_workflow(self) -> Workflow:
        return Workflow.from_workflow_data(self.workflow)



from enum import Enum
from pydantic import BaseModel, Field
from typing import List, Optional, Dict, Any, AsyncGenerator
import uuid
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
import re
import logging
from botrun_flow_lang.models.nodes.event import NodeEvent


class NodeType(str, Enum):
    START = "start"
    END = "end"
    ANSWER = "answer"
    LLM = "llm"
    CODE = "code"
    HTTP_REQUEST = "http-request"
    ITERATION = "iteration"
    SEARCH_AND_SCRAPE = "search-and-scrape"
    PERPLEXITY = "perplexity"
    VERTEX_AI_SEARCH = "vertex_ai_search"


class BaseNodeData(BaseModel):
    """
    @param printed_output: 可以指定一個 output_variables 裡面的變數名稱，在 NodeRunCompletedEvent 的時候會印出來
    """

    type: NodeType
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    title: str
    desc: Optional[str] = ""
    input_variables: List[InputVariable] = Field(default_factory=list)
    output_variables: List[OutputVariable] = Field(default_factory=list)
    print_start: bool = False
    print_stream: bool = False
    print_complete: bool = False
    complete_output: Optional[str] = None


class BaseNode(BaseModel):
    data: BaseNodeData

    async def run(
        self,
        variable_pool: Dict[str, Dict[str, Any]],
    ) -> AsyncGenerator[NodeEvent, None]:
        raise NotImplementedError("This method should be implemented by subclasses")

    def get_variable(
        self, variable_pool: Dict[str, Dict[str, Any]], node_id: str, variable_name: str
    ) -> Any:
        if node_id in variable_pool and variable_name in variable_pool[node_id]:
            return variable_pool[node_id][variable_name]
        return None

    def update_variable_pool(
        self, variable_pool: Dict[str, Dict[str, Any]], node_output: Dict[str, Any]
    ):
        if self.data.id not in variable_pool:
            variable_pool[self.data.id] = {}
        variable_pool[self.data.id].update(node_output)

    def replace_variables(
        self, text: str, variable_pool: Dict[str, Dict[str, Any]]
    ) -> str:
        pattern = r"{{#([\w.-]+)#}}"
        matches = re.finditer(pattern, text)

        for match in matches:
            full_match = match.group(0)
            try:
                node_id, var_name = match.group(1).rsplit(".", 1)
                if node_id in variable_pool and var_name in variable_pool[node_id]:
                    replacement = str(variable_pool[node_id][var_name])
                    text = text.replace(full_match, replacement)
                else:
                    logging.warning(f"Variable not found in pool: {full_match}")
            except ValueError:
                logging.error(f"Invalid variable format: {full_match}")

        return text



from fastapi import FastAPI
from botrun_flow_lang.api.routes import router
from fastapi.middleware.cors import CORSMiddleware

app = FastAPI()

app.include_router(router)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # 允許所有來源
    allow_credentials=True,
    allow_methods=["*"],  # 允許所有方法
    allow_headers=["*"],  # 允許所有頭
)


@app.get("/heartbeat")
async def heartbeat():
    return {"status": "ok"}


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)



import asyncio
from copy import deepcopy
import time
from typing import List, Dict, Any, AsyncGenerator
from botrun_flow_lang.models.nodes.base_node import BaseNode, NodeType
from botrun_flow_lang.models.workflow import Workflow, WorkflowItem
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunStartedEvent,
    NodeRunCompletedEvent,
    NodeRunStreamEvent,
    WorkflowRunStartedEvent,
    WorkflowRunCompletedEvent,
    WorkflowRunFailedEvent,
)


class WorkflowEngine:
    def __init__(self, workflow: Workflow):
        self.workflow = workflow
        self.variable_pool: Dict[str, Dict[str, Any]] = {}
        self.execution_times = []

    async def execute(
        self, initial_inputs: Dict[str, Any]
    ) -> AsyncGenerator[NodeEvent, None]:
        yield WorkflowRunStartedEvent()
        self.variable_pool.update(initial_inputs)

        try:
            async for event in self._execute_items(self.workflow.items):
                yield event

            yield WorkflowRunCompletedEvent(
                outputs=self.variable_pool, execution_times=self.execution_times
            )
        except Exception as e:
            import traceback

            traceback.print_exc()
            yield WorkflowRunFailedEvent(error=str(e))

    async def _execute_items(
        self, items: List[WorkflowItem], variable_pool: Dict[str, Dict[str, Any]] = None
    ) -> AsyncGenerator[NodeEvent, None]:
        var_pool = variable_pool if variable_pool is not None else self.variable_pool

        for item in items:
            if item.node:
                time_start = time.time()
                yield NodeRunStartedEvent(
                    node_id=item.node.data.id,
                    node_title=item.node.data.title,
                    node_type=item.node.data.type.value,
                    is_print=item.node.data.print_start,
                )

                if item.node.data.type == NodeType.ITERATION:
                    async for event in self._execute_iteration(item.node, item.items):
                        yield event
                else:
                    async for event in item.node.run(var_pool):
                        if isinstance(event, NodeRunStreamEvent):
                            yield event
                        elif isinstance(event, NodeRunCompletedEvent):
                            item.node.update_variable_pool(var_pool, event.outputs)
                            time_end = time.time()
                            self.execution_times.append(
                                f"=======>Node {item.node.data.title} completed in {time_end - time_start} seconds"
                            )
                            yield event
                        else:
                            yield event

    async def _execute_iteration(
        self, iteration_node: BaseNode, sub_items: List[WorkflowItem]
    ) -> AsyncGenerator[NodeEvent, None]:
        input_list = iteration_node.get_variable(
            self.variable_pool,
            iteration_node.data.input_selector.node_id,
            iteration_node.data.input_selector.variable_name,
        )

        if not isinstance(input_list, list):
            raise ValueError(
                f"Input for IterationNode must be a list, got {type(input_list)}"
            )

        outputs = []
        is_async = iteration_node.data.is_async

        if is_async:
            tasks = []
            for index, item in enumerate(input_list):
                task_variable_pool = {}
                for key, value in self.variable_pool.items():
                    if isinstance(value, dict):
                        task_variable_pool[key] = value.copy()
                    else:
                        task_variable_pool[key] = value

                task_variable_pool[iteration_node.data.id] = {
                    "item": item,
                    "index": index,
                }

                async def create_task(var_pool, current_index, current_item):
                    events = []
                    sub_engine = WorkflowEngine(self.workflow)
                    sub_engine.variable_pool = var_pool

                    async for event in sub_engine._execute_items(sub_items):
                        events.append(event)
                        yield event

                    if iteration_node.data.output_selector.node_id:
                        output = iteration_node.get_variable(
                            var_pool,
                            iteration_node.data.output_selector.node_id,
                            iteration_node.data.output_selector.variable_name,
                        )
                        outputs.append(output)

                    yield NodeRunStreamEvent(
                        node_id=iteration_node.data.id,
                        node_title=iteration_node.data.title,
                        node_type=iteration_node.data.type.value,
                        chunk=f"Iteration {current_index + 1}/{len(input_list)} completed",
                        is_print=iteration_node.data.print_stream,
                    )

                tasks.append(
                    self._process_item_wrapper(
                        create_task(task_variable_pool, index, item)
                    )
                )

            completed_tasks = await asyncio.gather(*tasks)
            for events in completed_tasks:
                for event in events:
                    yield event
        else:
            for index, item in enumerate(input_list):
                self.variable_pool[iteration_node.data.id] = {
                    "item": item,
                    "index": index,
                }
                async for event in self._execute_items(sub_items):
                    yield event

                if iteration_node.data.output_selector.node_id:
                    output = iteration_node.get_variable(
                        self.variable_pool,
                        iteration_node.data.output_selector.node_id,
                        iteration_node.data.output_selector.variable_name,
                    )
                    outputs.append(output)

        if outputs:
            iteration_node.update_variable_pool(self.variable_pool, {"output": outputs})
            yield NodeRunCompletedEvent(
                node_id=iteration_node.data.id,
                node_title=iteration_node.data.title,
                node_type=iteration_node.data.type.value,
                outputs={"output": outputs},
                complete_output=iteration_node.data.complete_output,
                is_print=iteration_node.data.print_complete,
            )

    async def _process_item_wrapper(self, generator):
        events = []
        async for event in generator:
            events.append(event)
        return events


async def run_workflow(
    workflow: Workflow, initial_inputs: Dict[str, Any]
) -> AsyncGenerator[NodeEvent, None]:
    engine = WorkflowEngine(workflow)
    async for event in engine.execute(initial_inputs):
        yield event



from pydantic import Field, field_validator
from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from typing import Dict, Any, List, AsyncGenerator

from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.nodes.event import NodeEvent, NodeRunCompletedEvent


class EndNodeData(BaseNodeData):
    type: NodeType = NodeType.END
    output_variables: List[OutputVariable] = [
        OutputVariable(variable_name="final_output")
    ]

    @field_validator("input_variables")
    def validate_input_variables(cls, v):
        assert len(v) == 1, "EndNode must have exactly 1 input variable"
        return v

    @field_validator("output_variables")
    def validate_output_variables(cls, v):
        assert len(v) == 1, "EndNode must have exactly 1 output variable"
        assert (
            v[0].variable_name == "final_output"
        ), "EndNode's output variable must be named 'final_output'"
        return v


class EndNode(BaseNode):
    data: EndNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        input_value = self.get_variable(
            variable_pool,
            self.data.input_variables[0].node_id,
            self.data.input_variables[0].variable_name,
        )
        yield NodeRunCompletedEvent(
            node_id=self.data.id,
            node_title=self.data.title,
            node_type=self.data.type.value,
            outputs={"final_output": input_value},
            complete_output=self.data.complete_output,
            is_print=self.data.print_complete,
        )



from typing import Union, List, Tuple
from google.cloud.exceptions import GoogleCloudError
from botrun_flow_lang.constants import HATCH_STORE_NAME
from botrun_flow_lang.services.hatch.hatch import Hatch
from botrun_flow_lang.services.base.firestore_base import FirestoreBase
from google.cloud import firestore


class HatchFsStore(FirestoreBase):
    def __init__(self, env_name: str):
        super().__init__(f"{env_name}-{HATCH_STORE_NAME}")

    async def get_hatch(self, item_id: str) -> Union[Hatch, None]:
        doc_ref = self.collection.document(item_id)
        doc = doc_ref.get()
        if doc.exists:
            data = doc.to_dict()
            return Hatch(**data)
        else:
            print(f">============Getting hatch {item_id} not exists")
            return None

    async def set_hatch(self, item: Hatch):
        try:
            doc_ref = self.collection.document(str(item.id))
            doc_ref.set(item.model_dump())
            return True, item
        except GoogleCloudError as e:
            print(f"Error setting hatch {item.id}: {e}")
            return False, None

    async def delete_hatch(self, item_id: str):
        try:
            doc_ref = self.collection.document(item_id)
            doc_ref.delete()
            return True
        except GoogleCloudError as e:
            print(f"Error deleting hatch {item_id}: {e}")
            return False

    async def get_hatches(
        self, user_id: str, offset: int = 0, limit: int = 20
    ) -> Tuple[List[Hatch], str]:
        try:
            query = (
                self.collection.where(
                    filter=firestore.FieldFilter("user_id", "==", user_id)
                )
                .order_by("name")
                .offset(offset)
                .limit(limit)
            )

            docs = query.stream()
            hatches = [Hatch(**doc.to_dict()) for doc in docs]
            return hatches, ""
        except GoogleCloudError as e:
            import traceback

            traceback.print_exc()
            print(f"Error getting hatches for user {user_id}: {e}")
            return [], f"Error getting hatches for user {user_id}: {e}"
        except Exception as e:
            import traceback

            traceback.print_exc()
            print(f"Error getting hatches for user {user_id}: {e}")
            return [], f"Error getting hatches for user {user_id}: {e}"

    async def get_default_hatch(self, user_id: str) -> Union[Hatch, None]:
        try:
            query = (
                self.collection.where(
                    filter=firestore.FieldFilter("user_id", "==", user_id)
                )
                .where(filter=firestore.FieldFilter("is_default", "==", True))
                .limit(1)
            )
            docs = query.stream()
            for doc in docs:
                return Hatch(**doc.to_dict())
            return None
        except GoogleCloudError as e:
            print(f"Error getting default hatch for user {user_id}: {e}")
            return None

    async def set_default_hatch(self, user_id: str, hatch_id: str) -> Tuple[bool, str]:
        try:
            # 获取当前的默认 hatch
            current_default = await self.get_default_hatch(user_id)

            # 获取要设置为默认的 hatch
            new_default = await self.get_hatch(hatch_id)
            if not new_default or new_default.user_id != user_id:
                return (
                    False,
                    f"Hatch with id {hatch_id} not found or does not belong to user {user_id}",
                )

            # 更新当前默认 hatch（如果存在）
            if current_default and current_default.id != hatch_id:
                current_default.is_default = False
                success, _ = await self.set_hatch(current_default)
                if not success:
                    return (
                        False,
                        f"Failed to update current default hatch {current_default.id}",
                    )

            # 设置新的默认 hatch
            new_default.is_default = True
            success, _ = await self.set_hatch(new_default)
            if not success:
                return False, f"Failed to set hatch {hatch_id} as default"

            return (
                True,
                f"Successfully set hatch {hatch_id} as default for user {user_id}",
            )
        except Exception as e:
            print(f"Error setting default hatch: {e}")
            return False, f"An error occurred: {str(e)}"



from typing import Dict, Any, AsyncGenerator, List
from pydantic import Field, field_validator

from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.variable import InputVariable
from botrun_flow_lang.models.nodes.event import NodeEvent


class IterationNodeData(BaseNodeData):
    type: NodeType = NodeType.ITERATION
    input_selector: InputVariable
    output_selector: InputVariable
    is_async: bool = True  # 新增字段，默认为 True

    @field_validator("input_selector")
    def validate_input_selector(cls, v):
        assert v is not None, "IterationNode must have an input_selector"
        return v

    @field_validator("output_selector")
    def validate_output_selector(cls, v):
        assert v is not None, "IterationNode must have an output_selector"
        return v


class IterationNode(BaseNode):
    data: IterationNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        # This method is now handled by WorkflowEngine._execute_iteration
        raise NotImplementedError("IterationNode.run should not be called directly")



from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunCompletedEvent,
)
from typing import Dict, Any, AsyncGenerator
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from pydantic import Field, field_validator


class AnswerNodeData(BaseNodeData):
    type: NodeType = NodeType.ANSWER
    input_variables: list[InputVariable] = Field(default_factory=list)
    output_variables: list[OutputVariable] = [OutputVariable(variable_name="answer")]

    @field_validator("output_variables")
    def validate_output_variables(cls, v):
        assert len(v) == 1, "AnswerNode must have exactly 1 output variable"
        assert (
            v[0].variable_name == "answer"
        ), "AnswerNode's output variable must be named 'answer'"
        return v


class AnswerNode(BaseNode):
    data: AnswerNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        output = variable_pool.get(self.data.input_variables[0].node_id, {}).get(
            self.data.input_variables[0].variable_name, ""
        )

        yield NodeRunCompletedEvent(
            node_id=self.data.id,
            node_title=self.data.title,
            node_type=self.data.type.value,
            outputs={"answer": output},
            complete_output=self.data.complete_output,
            is_print=self.data.print_complete,
        )



from pydantic import BaseModel
import yaml
from enum import Enum


class BotrunAppMode(str, Enum):
    CHATBOT = "chatbot"
    WORKFLOW = "workflow"


class BotrunApp(BaseModel):
    name: str
    description: str
    mode: BotrunAppMode

    def to_yaml(self) -> str:
        return yaml.dump(self.model_dump(mode="json"))

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "BotrunApp":
        data = yaml.safe_load(yaml_str)
        return cls(**data)



from pydantic import BaseModel
from typing import Optional


class LlmAgent(BaseModel):
    """
    @include_in_history: 是否將這次的回答加入 history
    @max_system_prompt_length: 如果 system prompt 超過這個長度，會整個被捨棄
    """

    name: str = ""
    print_output: bool = True
    print_plotly: bool = True
    output: Optional[str] = None
    model: str = "openai/gpt-4o-2024-08-06"
    system_prompt: str
    gen_image: bool = False
    include_in_history: bool = True
    max_system_prompt_length: Optional[int] = None



from typing import Any, AsyncGenerator, Dict, List, Optional
from pydantic import Field, field_validator, BaseModel
from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunStreamEvent,
    NodeRunCompletedEvent,
    NodeRunFailedEvent,
)
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
import aiohttp
import os
import json
from dotenv import load_dotenv

load_dotenv()


class PerplexityModelConfig(BaseModel):
    completion_params: Dict[str, Any] = Field(default_factory=dict)
    name: str = "llama-3.1-sonar-huge-128k-online"


class PerplexityNodeData(BaseNodeData):
    """
    @param input_variables: 輸入變數，可以傳入 prompt 和 history
    @param model: 模型設定，包含模型名稱和完成參數
    @param temperature: 控制隨機性的參數，介於 0-2 之間
    @param search_recency_filter: 搜尋結果的時間範圍，可選值為 month、week、day、hour
    @param prompt_template: 提示詞模板，格式為 List[Dict[str, str]]，每個 dict 包含 role 和 content
    """

    type: NodeType = NodeType.PERPLEXITY
    model: PerplexityModelConfig = Field(default_factory=PerplexityModelConfig)
    prompt_template: List[Dict[str, str]] = []
    temperature: float = Field(default=0.2, ge=0, lt=2)
    # search_recency_filter: str = "day"
    input_variables: List[InputVariable] = []
    output_variables: List[OutputVariable] = [
        OutputVariable(variable_name="perplexity_output")
    ]
    print_stream: bool = True

    @field_validator("output_variables")
    def validate_output_variables(cls, v):
        assert len(v) == 1, "PerplexityNode must have exactly 1 output variable"
        assert (
            v[0].variable_name == "perplexity_output"
        ), "PerplexityNode's output variable must be named 'perplexity_output'"
        return v


PERPLEXITY_API_URL = "https://api.perplexity.ai/chat/completions"


class PerplexityNode(BaseNode):
    data: PerplexityNodeData

    def _get_api_key(self) -> str:
        api_key = os.getenv("PPLX_API_KEY")
        if not api_key:
            raise ValueError("PPLX_API_KEY environment variable not set")
        return api_key

    def prepare_messages(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> List[Dict[str, str]]:
        messages = []

        # Add messages from prompt template
        for message in self.data.prompt_template:
            content = self.replace_variables(message["content"], variable_pool)
            messages.append({"role": message["role"], "content": content})

        # Add history if provided
        history_var = next(
            (
                var
                for var in self.data.input_variables
                if var.variable_name == "history"
            ),
            None,
        )
        if history_var:
            history = self.get_variable(variable_pool, history_var.node_id, "history")
            if isinstance(history, list):
                if len(history) > 0:
                    if history[-1]["role"] == "user":
                        history.pop()
                history.extend(messages)
                messages = history

        # Add prompt if provided
        prompt_var = next(
            (var for var in self.data.input_variables if var.variable_name == "prompt"),
            None,
        )
        if prompt_var:
            prompt = self.get_variable(variable_pool, prompt_var.node_id, "prompt")
            if isinstance(prompt, str):
                messages.append({"role": "user", "content": prompt})

        return messages

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        try:
            messages = self.prepare_messages(variable_pool)
            if not messages:
                raise ValueError("No messages provided")

            headers = {
                "Authorization": f"Bearer {self._get_api_key()}",
                "Content-Type": "application/json",
            }

            payload = {
                "model": self.data.model.name,
                "messages": messages,
                "temperature": self.data.temperature,
                "stream": True,
                # "search_recency_filter": self.data.search_recency_filter,
                **self.data.model.completion_params,
            }

            async with aiohttp.ClientSession() as session:
                async with session.post(
                    PERPLEXITY_API_URL, headers=headers, json=payload
                ) as response:
                    if response.status != 200:
                        error_text = await response.text()
                        raise ValueError(f"Perplexity API error: {error_text}")

                    full_response = ""
                    async for line in response.content:
                        if line:
                            line = line.decode("utf-8").strip()
                            if line.startswith("data: "):
                                line = line[6:]  # Remove 'data: ' prefix
                                if line == "[DONE]":
                                    break

                                try:
                                    chunk_data = json.loads(line)
                                    if (
                                        chunk_data["choices"][0]
                                        .get("delta", {})
                                        .get("content")
                                    ):
                                        content = chunk_data["choices"][0]["delta"][
                                            "content"
                                        ]
                                        full_response += content
                                        yield NodeRunStreamEvent(
                                            node_id=self.data.id,
                                            node_title=self.data.title,
                                            node_type=self.data.type.value,
                                            chunk=content,
                                            is_print=self.data.print_stream,
                                        )
                                except json.JSONDecodeError:
                                    continue

            yield NodeRunCompletedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                outputs={"perplexity_output": full_response},
                complete_output=self.data.complete_output,
                is_print=self.data.print_complete,
            )

        except Exception as e:
            yield NodeRunFailedEvent(
                node_id=self.data.id,
                node_title=self.data.title,
                node_type=self.data.type.value,
                error=str(e),
                is_print=True,
            )
            raise



import unittest
from botrun_flow_lang.models.nodes.start_node import StartNodeData
from botrun_flow_lang.models.workflow_config import WorkflowConfig
from botrun_flow_lang.models.botrun_app import BotrunApp, BotrunAppMode
from botrun_flow_lang.models.workflow import WorkflowData
from botrun_flow_lang.models.nodes.llm_node import LLMNodeData, LLMModelConfig
from botrun_flow_lang.utils.yaml_utils import compare_yaml_with_deepdiff
from botrun_flow_lang.models.nodes.base_node import NodeType
from botrun_flow_lang.models.nodes.end_node import EndNodeData
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.nodes.answer_node import AnswerNodeData


class TestWorkflowConfig(unittest.TestCase):
    def test_to_yaml(self):
        botrun_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.CHATBOT,
        )
        model_config = LLMModelConfig(
            completion_params={
                "max_tokens": 512,
                "temperature": 0.7,
            },
            mode="chat",
            name="gpt-3.5-turbo",
            provider="openai",
        )
        llm_node = LLMNodeData(
            title="Test LLM",
            id="idLLMNodeData",
            model=model_config,
            prompt_template=[
                {"role": "system", "text": "You are a helpful assistant."}
            ],
            input_variables=[],
            output_variables=[OutputVariable(variable_name="test_var")],
        )
        answer_node = AnswerNodeData(
            title="Answer Node",
            id="idAnswerNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="test_var")
            ],
        )
        workflow = WorkflowData(nodes=[llm_node, answer_node])
        workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
        yaml_str = workflow_config.to_yaml()
        expected_yaml = """
botrun_app:
  name: Test App
  description: A test application
  mode: chatbot
workflow:
  nodes:
    - type: llm
      id: idLLMNodeData
      title: Test LLM
      desc: ''
      model:
        completion_params:
          max_tokens: 512
          temperature: 0.7
        mode: chat
        name: gpt-3.5-turbo
        provider: openai
      prompt_template:
        - role: system
          text: You are a helpful assistant.
      context: {}
      input_variables: []
      output_variables:
        - variable_name: test_var
      vision: {}
    - type: answer
      id: idAnswerNodeData
      title: Answer Node
      desc: ''
      input_variables:
        - node_id: idLLMNodeData
          variable_name: test_var
      output_variables:
        - variable_name: answer
"""
        diff = compare_yaml_with_deepdiff(yaml_str, expected_yaml)
        self.assertEqual(diff, {})

    def test_from_yaml(self):
        yaml_str = """
botrun_app:
  name: Test App
  description: A test application
  mode: workflow
workflow:
  nodes:
    - type: llm
      id: idLLMNodeData
      title: Test LLM
      model:
        completion_params:
          max_tokens: 512
          temperature: 0.7
        mode: chat
        name: gpt-3.5-turbo
        provider: openai
      prompt_template:
        - role: system
          text: You are a helpful assistant.
      context: {}
      input_variables: []
      output_variables:
        - variable_name: test_var
      vision: {}
    - type: end
      id: idEndNodeData
      title: End Node
      desc: ''
      input_variables:
        - node_id: idLLMNodeData
          variable_name: test_var
      output_variables:
        - variable_name: final_output
"""
        workflow_config = WorkflowConfig.from_yaml(yaml_str)
        self.assertEqual(workflow_config.botrun_app.name, "Test App")
        self.assertEqual(workflow_config.botrun_app.description, "A test application")
        self.assertEqual(workflow_config.botrun_app.mode, BotrunAppMode.WORKFLOW)
        self.assertEqual(len(workflow_config.workflow.nodes), 2)
        self.assertIsInstance(workflow_config.workflow.nodes[0], LLMNodeData)
        self.assertEqual(workflow_config.workflow.nodes[0].title, "Test LLM")
        self.assertEqual(workflow_config.workflow.nodes[0].id, "idLLMNodeData")
        self.assertEqual(len(workflow_config.workflow.nodes[0].output_variables), 1)
        self.assertEqual(
            workflow_config.workflow.nodes[0].output_variables[0].variable_name,
            "test_var",
        )

    def test_roundtrip(self):
        botrun_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.WORKFLOW,
        )
        model_config = LLMModelConfig(
            completion_params={
                "max_tokens": 512,
                "temperature": 0.7,
            },
            mode="chat",
            name="gpt-3.5-turbo",
            provider="openai",
        )
        llm_node = LLMNodeData(
            title="Test LLM",
            id="idLLMNodeData",
            model=model_config,
            prompt_template=[
                {"role": "system", "text": "You are a helpful assistant."}
            ],
            input_variables=[],
            output_variables=[OutputVariable(variable_name="test_var")],
        )
        end_node = EndNodeData(
            title="End Node",
            id="idEndNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="test_var")
            ],
        )
        workflow = WorkflowData(nodes=[llm_node, end_node])
        original_workflow_config = WorkflowConfig(
            botrun_app=botrun_app, workflow=workflow
        )
        yaml_str = original_workflow_config.to_yaml()
        reconstructed_workflow_config = WorkflowConfig.from_yaml(yaml_str)
        self.assertEqual(original_workflow_config, reconstructed_workflow_config)

    def test_workflow_with_start_and_llm_nodes(self):
        botrun_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.WORKFLOW,
        )

        start_node = StartNodeData(
            title="Start Node",
            id="idStartNodeData",
            input_variables=[],
            output_variables=[OutputVariable(variable_name="user_input")],
        )

        model_config = LLMModelConfig(
            completion_params={
                "max_tokens": 512,
                "temperature": 0.7,
            },
            mode="chat",
            name="gpt-3.5-turbo",
            provider="openai",
        )
        llm_node = LLMNodeData(
            title="Test LLM",
            id="idLLMNodeData",
            model=model_config,
            prompt_template=[
                {"role": "system", "text": "You are a helpful assistant."}
            ],
            input_variables=[
                InputVariable(node_id="idStartNodeData", variable_name="user_input")
            ],
            output_variables=[OutputVariable(variable_name="test_var")],
        )
        end_node = EndNodeData(
            title="End Node",
            id="idEndNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="test_var")
            ],
        )

        workflow = WorkflowData(nodes=[start_node, llm_node, end_node])
        workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)

        yaml_str = workflow_config.to_yaml()
        expected_yaml = """
botrun_app:
  name: Test App
  description: A test application
  mode: workflow
workflow:
  nodes:
    - type: start
      id: idStartNodeData
      title: Start Node
      desc: ''
      input_variables: []
      output_variables:
        - variable_name: user_input
    - type: llm
      id: idLLMNodeData
      title: Test LLM
      desc: ''
      model:
        completion_params:
          max_tokens: 512
          temperature: 0.7
        mode: chat
        name: gpt-3.5-turbo
        provider: openai
      prompt_template:
        - role: system
          text: You are a helpful assistant.
      context: {}
      input_variables:
        - node_id: idStartNodeData
          variable_name: user_input
      output_variables:
        - variable_name: test_var
      vision: {}
    - type: end
      id: idEndNodeData
      title: End Node
      desc: ''
      input_variables:
        - node_id: idLLMNodeData
          variable_name: test_var
      output_variables:
        - variable_name: final_output
"""
        diff = compare_yaml_with_deepdiff(yaml_str, expected_yaml)
        self.assertEqual(diff, {})

        # Test from_yaml
        reconstructed_workflow_config = WorkflowConfig.from_yaml(yaml_str)
        self.assertEqual(len(reconstructed_workflow_config.workflow.nodes), 3)
        self.assertIsInstance(
            reconstructed_workflow_config.workflow.nodes[0], StartNodeData
        )
        self.assertIsInstance(
            reconstructed_workflow_config.workflow.nodes[1], LLMNodeData
        )
        self.assertIsInstance(
            reconstructed_workflow_config.workflow.nodes[2], EndNodeData
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].title, "Test LLM"
        )

        # 在测试 from_yaml 时，检查 type 字段
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[0].type, NodeType.START
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].type, NodeType.LLM
        )

        # Additional from_yaml test
        self.assertEqual(reconstructed_workflow_config.botrun_app.name, "Test App")
        self.assertEqual(
            reconstructed_workflow_config.botrun_app.description, "A test application"
        )
        self.assertEqual(
            reconstructed_workflow_config.botrun_app.mode, BotrunAppMode.WORKFLOW
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[0].id, "idStartNodeData"
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].id, "idLLMNodeData"
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].model.name, "gpt-3.5-turbo"
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].model.provider, "openai"
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].model.completion_params[
                "max_tokens"
            ],
            512,
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].model.completion_params[
                "temperature"
            ],
            0.7,
        )

    def test_workflow_with_start_llm_and_end_nodes(self):
        botrun_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.WORKFLOW,
        )

        start_node = StartNodeData(
            title="Start Node",
            id="idStartNodeData",
            input_variables=[],
            output_variables=[OutputVariable(variable_name="user_input")],
        )

        model_config = LLMModelConfig(
            completion_params={
                "max_tokens": 512,
                "temperature": 0.7,
            },
            mode="chat",
            name="gpt-3.5-turbo",
            provider="openai",
        )
        llm_node = LLMNodeData(
            title="Test LLM",
            id="idLLMNodeData",
            model=model_config,
            prompt_template=[
                {"role": "system", "text": "You are a helpful assistant."}
            ],
            input_variables=[
                InputVariable(node_id="idStartNodeData", variable_name="user_input")
            ],
            output_variables=[OutputVariable(variable_name="llm_output")],
        )

        end_node = EndNodeData(
            title="End Node",
            id="idEndNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="llm_output")
            ],
        )

        workflow = WorkflowData(nodes=[start_node, llm_node, end_node])
        workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)

        yaml_str = workflow_config.to_yaml()
        expected_yaml = """
botrun_app:
  name: Test App
  description: A test application
  mode: workflow
workflow:
  nodes:
    - type: start
      id: idStartNodeData
      title: Start Node
      desc: ''
      input_variables: []
      output_variables:
        - variable_name: user_input
    - type: llm
      id: idLLMNodeData
      title: Test LLM
      desc: ''
      model:
        completion_params:
          max_tokens: 512
          temperature: 0.7
        mode: chat
        name: gpt-3.5-turbo
        provider: openai
      prompt_template:
        - role: system
          text: You are a helpful assistant.
      context: {}
      input_variables:
        - node_id: idStartNodeData
          variable_name: user_input
      output_variables:
        - variable_name: llm_output
      vision: {}
    - type: end
      id: idEndNodeData
      title: End Node
      desc: ''
      input_variables:
        - node_id: idLLMNodeData
          variable_name: llm_output
      output_variables:
        - variable_name: final_output
"""
        diff = compare_yaml_with_deepdiff(yaml_str, expected_yaml)
        self.assertEqual(diff, {})

        # Test from_yaml
        reconstructed_workflow_config = WorkflowConfig.from_yaml(yaml_str)
        self.assertEqual(len(reconstructed_workflow_config.workflow.nodes), 3)
        self.assertIsInstance(
            reconstructed_workflow_config.workflow.nodes[0], StartNodeData
        )
        self.assertIsInstance(
            reconstructed_workflow_config.workflow.nodes[1], LLMNodeData
        )
        self.assertIsInstance(
            reconstructed_workflow_config.workflow.nodes[2], EndNodeData
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[0].type, NodeType.START
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1].type, NodeType.LLM
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[2].type, NodeType.END
        )
        self.assertEqual(
            reconstructed_workflow_config.botrun_app.mode, BotrunAppMode.WORKFLOW
        )

        # Check variables
        self.assertEqual(
            len(reconstructed_workflow_config.workflow.nodes[0].output_variables), 1
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[0]
            .output_variables[0]
            .variable_name,
            "user_input",
        )

        self.assertEqual(
            len(reconstructed_workflow_config.workflow.nodes[1].input_variables), 1
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1]
            .input_variables[0]
            .variable_name,
            "user_input",
        )
        self.assertEqual(
            len(reconstructed_workflow_config.workflow.nodes[1].output_variables), 1
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[1]
            .output_variables[0]
            .variable_name,
            "llm_output",
        )

        self.assertEqual(
            len(reconstructed_workflow_config.workflow.nodes[2].input_variables), 1
        )
        self.assertEqual(
            reconstructed_workflow_config.workflow.nodes[2]
            .input_variables[0]
            .variable_name,
            "llm_output",
        )

    def test_workflow_mode_validation(self):
        botrun_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.WORKFLOW,
        )

        start_node = StartNodeData(
            title="Start Node",
            id="idStartNodeData",
            input_variables=[],
            output_variables=[OutputVariable(variable_name="user_input")],
        )

        llm_node = LLMNodeData(
            title="Test LLM",
            id="idLLMNodeData",
            model=LLMModelConfig(
                completion_params={
                    "max_tokens": 512,
                    "temperature": 0.7,
                },
                mode="chat",
                name="gpt-3.5-turbo",
                provider="openai",
            ),
            prompt_template=[
                {"role": "system", "text": "You are a helpful assistant."}
            ],
            input_variables=[
                InputVariable(node_id="idStartNodeData", variable_name="user_input")
            ],
            output_variables=[OutputVariable(variable_name="llm_output")],
        )

        end_node = EndNodeData(
            title="End Node",
            id="idEndNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="llm_output")
            ],
        )

        # This should work
        workflow = WorkflowData(nodes=[start_node, llm_node, end_node])
        WorkflowConfig(botrun_app=botrun_app, workflow=workflow)

        # This should fail
        answer_node = AnswerNodeData(
            title="Answer Node",
            id="idAnswerNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="llm_output")
            ],
        )
        workflow_invalid = WorkflowData(nodes=[start_node, llm_node, answer_node])
        with self.assertRaises(ValueError):
            WorkflowConfig(botrun_app=botrun_app, workflow=workflow_invalid)

    def test_chatbot_mode_validation(self):
        botrun_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.CHATBOT,
        )

        start_node = StartNodeData(
            title="Start Node",
            id="idStartNodeData",
            input_variables=[],
            output_variables=[OutputVariable(variable_name="user_input")],
        )

        llm_node = LLMNodeData(
            title="Test LLM",
            id="idLLMNodeData",
            model=LLMModelConfig(
                completion_params={
                    "max_tokens": 512,
                    "temperature": 0.7,
                },
                mode="chat",
                name="gpt-3.5-turbo",
                provider="openai",
            ),
            prompt_template=[
                {"role": "system", "text": "You are a helpful assistant."}
            ],
            input_variables=[
                InputVariable(node_id="idStartNodeData", variable_name="user_input")
            ],
            output_variables=[OutputVariable(variable_name="llm_output")],
        )

        answer_node = AnswerNodeData(
            title="Answer Node",
            id="idAnswerNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="llm_output")
            ],
        )

        # This should work
        workflow = WorkflowData(nodes=[start_node, llm_node, answer_node])
        WorkflowConfig(botrun_app=botrun_app, workflow=workflow)

        # This should fail
        end_node = EndNodeData(
            title="End Node",
            id="idEndNodeData",
            input_variables=[
                InputVariable(node_id="idLLMNodeData", variable_name="llm_output")
            ],
        )
        workflow_invalid = WorkflowData(nodes=[start_node, llm_node, end_node])
        with self.assertRaises(ValueError):
            WorkflowConfig(botrun_app=botrun_app, workflow=workflow_invalid)


if __name__ == "__main__":
    unittest.main()



import unittest
import asyncio
from botrun_flow_lang.models.nodes.base_node import NodeType
from botrun_flow_lang.models.nodes.answer_node import AnswerNode, AnswerNodeData
from botrun_flow_lang.models.nodes.event import (
    NodeRunStreamEvent,
    NodeRunCompletedEvent,
)
from botrun_flow_lang.models.variable import InputVariable
import uuid


class TestAnswerNode(unittest.TestCase):
    def test_answer_node_data_creation(self):
        answer_node = AnswerNodeData(title="Test Answer")

        self.assertEqual(answer_node.type, NodeType.ANSWER)
        self.assertEqual(answer_node.title, "Test Answer")

        # Test that id is a string and can be parsed as a valid UUID
        self.assertIsInstance(answer_node.id, str)
        self.assertTrue(uuid.UUID(answer_node.id, version=4))

    async def async_test_answer_node_run(self):
        answer_node = AnswerNode(
            data=AnswerNodeData(
                title="Test Answer",
                input_variables=[
                    InputVariable(node_id="idLLMNodeData", variable_name="llm_output")
                ],
            )
        )
        variable_pool = {"idLLMNodeData": {"llm_output": "Test output"}}

        async for event in answer_node.run(variable_pool):
            if isinstance(event, NodeRunStreamEvent):
                self.assertEqual(event.chunk, "Test output")
            elif isinstance(event, NodeRunCompletedEvent):
                self.assertEqual(event.outputs, {"answer": "Test output"})

    def test_answer_node_run(self):
        asyncio.run(self.async_test_answer_node_run())


if __name__ == "__main__":
    unittest.main()



import unittest
from botrun_flow_lang.models.botrun_app import BotrunApp, BotrunAppMode


class TestBotrunApp(unittest.TestCase):
    def test_to_yaml(self):
        app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.CHATBOT,
        )
        yaml_str = app.to_yaml()
        self.assertIn("name: Test App", yaml_str)
        self.assertIn("description: A test application", yaml_str)
        self.assertIn("mode: chatbot", yaml_str)

    def test_from_yaml(self):
        yaml_str = """
        name: Test App
        description: A test application
        mode: workflow
        """
        app = BotrunApp.from_yaml(yaml_str)
        self.assertEqual(app.name, "Test App")
        self.assertEqual(app.description, "A test application")
        self.assertEqual(app.mode, BotrunAppMode.WORKFLOW)

    def test_roundtrip(self):
        original_app = BotrunApp(
            name="Test App",
            description="A test application",
            mode=BotrunAppMode.CHATBOT,
        )
        yaml_str = original_app.to_yaml()
        reconstructed_app = BotrunApp.from_yaml(yaml_str)
        self.assertEqual(original_app, reconstructed_app)

    def test_invalid_mode(self):
        with self.assertRaises(ValueError):
            BotrunApp(
                name="Test App", description="A test application", mode="invalid_mode"
            )


if __name__ == "__main__":
    unittest.main()



from deepdiff import DeepDiff
import yaml


def compare_yaml_with_deepdiff(yaml_str1, yaml_str2):
    dict1 = yaml.safe_load(yaml_str1)
    dict2 = yaml.safe_load(yaml_str2)
    diff = DeepDiff(dict1, dict2, ignore_order=True)
    return diff



from fastapi import APIRouter, HTTPException, Depends, Query
from typing import List
from botrun_flow_lang.services.user_workflow.user_workflow import UserWorkflow
from botrun_flow_lang.services.user_workflow.user_workflow_factory import (
    user_workflow_store_factory,
)
from botrun_flow_lang.services.user_workflow.user_workflow_fs_store import (
    UserWorkflowFsStore,
)

router = APIRouter()


async def get_user_workflow_store():
    return user_workflow_store_factory()


@router.post("/user_workflow", response_model=UserWorkflow)
async def create_user_workflow(
    user_workflow: UserWorkflow,
    store: UserWorkflowFsStore = Depends(get_user_workflow_store),
):
    success, created_workflow = await store.set_user_workflow(user_workflow)
    if not success:
        raise HTTPException(status_code=500, detail="Failed to create user workflow")
    return created_workflow


@router.put("/user_workflow/{workflow_id}", response_model=UserWorkflow)
async def update_user_workflow(
    workflow_id: str,
    user_workflow: UserWorkflow,
    store: UserWorkflowFsStore = Depends(get_user_workflow_store),
):
    existing_workflow = await store.get_user_workflow(workflow_id)
    if not existing_workflow:
        raise HTTPException(status_code=404, detail="User workflow not found")
    user_workflow.id = workflow_id
    success, updated_workflow = await store.set_user_workflow(user_workflow)
    if not success:
        raise HTTPException(status_code=500, detail="Failed to update user workflow")
    return updated_workflow


@router.delete("/user_workflow/{workflow_id}")
async def delete_user_workflow(
    workflow_id: str, store: UserWorkflowFsStore = Depends(get_user_workflow_store)
):
    success = await store.delete_user_workflow(workflow_id)
    if not success:
        raise HTTPException(
            status_code=500,
            detail={
                "success": False,
                "message": f"Failed to delete user workflow {workflow_id}",
            },
        )
    return {
        "success": True,
        "message": f"User workflow {workflow_id} deleted successfully",
    }


@router.get("/user_workflow/{workflow_id}", response_model=UserWorkflow)
async def get_user_workflow(
    workflow_id: str, store: UserWorkflowFsStore = Depends(get_user_workflow_store)
):
    user_workflow = await store.get_user_workflow(workflow_id)
    if not user_workflow:
        raise HTTPException(status_code=404, detail="User workflow not found")
    return user_workflow


@router.get("/user_workflows", response_model=List[UserWorkflow])
async def get_user_workflows(
    user_id: str,
    offset: int = Query(0, ge=0),
    limit: int = Query(20, ge=1, le=20),
    store: UserWorkflowFsStore = Depends(get_user_workflow_store),
):
    workflows, error = await store.get_user_workflows(user_id, offset, limit)
    if error:
        raise HTTPException(status_code=500, detail=error)
    return workflows



from pydantic import BaseModel
from typing import List, Union, Optional
import yaml
from botrun_flow_lang.models.nodes.base_node import BaseNode, NodeType
from botrun_flow_lang.models.nodes.code_node import CodeNode, CodeNodeData
from botrun_flow_lang.models.nodes.http_request_node import (
    HttpRequestNode,
    HttpRequestNodeData,
)
from botrun_flow_lang.models.nodes.llm_node import LLMNodeData, LLMNode
from botrun_flow_lang.models.nodes.perplexity_node import (
    PerplexityNode,
    PerplexityNodeData,
)
from botrun_flow_lang.models.nodes.start_node import StartNodeData, StartNode
from botrun_flow_lang.models.nodes.end_node import EndNodeData, EndNode
from botrun_flow_lang.models.nodes.answer_node import AnswerNodeData, AnswerNode
from botrun_flow_lang.models.nodes.iteration_node import (
    IterationNodeData,
    IterationNode,
)
from botrun_flow_lang.models.nodes.search_and_scrape_node import (
    SearchAndScrapeNodeData,
    SearchAndScrapeNode,
)
from botrun_flow_lang.models.nodes.vertex_ai_search_node import (
    VertexAiSearchNode,
    VertexAiSearchNodeData,
)

NodeData = Union[
    LLMNodeData,
    StartNodeData,
    EndNodeData,
    AnswerNodeData,
    CodeNodeData,
    HttpRequestNodeData,
    IterationNodeData,
    SearchAndScrapeNodeData,
    PerplexityNodeData,
    VertexAiSearchNodeData,
]
Node = Union[
    LLMNode,
    StartNode,
    EndNode,
    AnswerNode,
    CodeNode,
    HttpRequestNode,
    IterationNode,
    SearchAndScrapeNode,
    PerplexityNode,
    VertexAiSearchNode,
]


class WorkflowItem(BaseModel):
    node: Optional[Node] = None
    items: Optional[List["WorkflowItem"]] = None


class Workflow(BaseModel):
    items: List[WorkflowItem]

    @classmethod
    def from_workflow_data(cls, workflow_data: "WorkflowData") -> "Workflow":
        return cls(items=cls._create_workflow_items(workflow_data.nodes))

    @classmethod
    def _create_workflow_items(
        cls, nodes: List[Union[NodeData, List]]
    ) -> List[WorkflowItem]:
        items = []
        for node in nodes:
            if isinstance(node, list):
                iteration_node = cls._create_node(node[0])
                sub_items = cls._create_workflow_items(node[1:])
                items.append(WorkflowItem(node=iteration_node, items=sub_items))
            else:
                items.append(WorkflowItem(node=cls._create_node(node)))
        return items

    @staticmethod
    def _create_node(node_data: NodeData) -> Node:
        if isinstance(node_data, StartNodeData):
            return StartNode(data=node_data)
        elif isinstance(node_data, LLMNodeData):
            return LLMNode(data=node_data)
        elif isinstance(node_data, EndNodeData):
            return EndNode(data=node_data)
        elif isinstance(node_data, AnswerNodeData):
            return AnswerNode(data=node_data)
        elif isinstance(node_data, CodeNodeData):
            return CodeNode(data=node_data)
        elif isinstance(node_data, HttpRequestNodeData):
            return HttpRequestNode(data=node_data)
        elif isinstance(node_data, IterationNodeData):
            return IterationNode(data=node_data)
        elif isinstance(node_data, SearchAndScrapeNodeData):
            return SearchAndScrapeNode(data=node_data)
        elif isinstance(node_data, PerplexityNodeData):
            return PerplexityNode(data=node_data)
        elif isinstance(node_data, VertexAiSearchNodeData):
            return VertexAiSearchNode(data=node_data)
        else:
            raise ValueError(f"Unknown node type: {type(node_data)}")


class WorkflowData(BaseModel):
    nodes: List[Union[NodeData, List[Union[NodeData, List]]]]

    def to_yaml(self) -> str:
        return yaml.dump(self.model_dump(mode="json"))

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "WorkflowData":
        data = yaml.safe_load(yaml_str)
        return cls(**data)



import unittest
from botrun_flow_lang.models.nodes.base_node import NodeType
from botrun_flow_lang.models.nodes.llm_node import LLMNodeData, LLMModelConfig
import uuid


class TestLLMNode(unittest.TestCase):
    def test_llm_node_data_creation(self):
        model_config = LLMModelConfig(
            completion_params={
                "frequency_penalty": 0,
                "max_tokens": 512,
                "presence_penalty": 0,
                "temperature": 0.7,
                "top_p": 1,
            },
            mode="chat",
            name="gpt-3.5-turbo",
            provider="openai",
        )

        llm_node = LLMNodeData(
            title="LLM 2",
            model=model_config,
            prompt_template=[
                {
                    "role": "system",
                    "text": "<Task>\nDo a general overview style summary to the following text. Use the same language as text to be summarized. \n<Text to be summarized>\n{{#1711526002155.input#}}\n<Summary>",
                }
            ],
            context={"enabled": False, "variable_selector": []},
            variables=[],
            vision={},
        )

        self.assertEqual(llm_node.type, NodeType.LLM)
        self.assertEqual(llm_node.title, "LLM 2")
        self.assertEqual(llm_node.model.name, "gpt-3.5-turbo")
        self.assertEqual(llm_node.model.provider, "openai")
        self.assertEqual(len(llm_node.prompt_template), 1)
        self.assertEqual(llm_node.context["enabled"], False)
        self.assertEqual(llm_node.input_variables, [])
        self.assertEqual(llm_node.output_variables, [])
        self.assertEqual(llm_node.vision, {})

        # Test that id is a string and can be parsed as a valid UUID
        self.assertIsInstance(llm_node.id, str)
        self.assertTrue(uuid.UUID(llm_node.id, version=4))


if __name__ == "__main__":
    unittest.main()



from pydantic import BaseModel, Field, field_validator
from typing import List, Dict, Any, AsyncGenerator
from botrun_flow_lang.models.nodes.base_node import BaseNode, BaseNodeData, NodeType
import litellm
from botrun_flow_lang.models.nodes.event import (
    NodeEvent,
    NodeRunStreamEvent,
    NodeRunCompletedEvent,
)

import os
from dotenv import load_dotenv

from botrun_flow_lang.models.variable import InputVariable, OutputVariable

load_dotenv()


class LLMModelConfig(BaseModel):
    completion_params: Dict[str, Any] = Field(default_factory=dict)
    name: str


class LLMNodeData(BaseNodeData):
    """
    @param input_variables: 輸入變數，目前只有 history 可以傳入，如果有輸入的話，回答時會帶入 history 的內容，也可以不傳入
    @param stream: 是否要以 stream 的方式返回訊息
    """

    type: NodeType = NodeType.LLM
    model: LLMModelConfig
    prompt_template: List[Dict[str, str]]
    context: Dict[str, Any] = Field(default_factory=dict)
    vision: Dict[str, bool] = Field(default_factory=dict)
    stream: bool = True
    print_stream: bool = True
    output_variables: List[OutputVariable] = [
        OutputVariable(variable_name="llm_output"),
    ]

    @field_validator("output_variables")
    def validate_output_variables(cls, v):
        assert len(v) == 1, "LLMNode must have exactly 1 output variable"
        assert (
            v[0].variable_name == "llm_output"
        ), "LLMNode's output variable must be named 'llm_output'"
        return v


class LLMNode(BaseNode):
    data: LLMNodeData

    async def run(
        self, variable_pool: Dict[str, Dict[str, Any]]
    ) -> AsyncGenerator[NodeEvent, None]:
        messages = self.prepare_messages(variable_pool)
        if len(self.data.input_variables) > 0:
            history = self.get_variable(
                variable_pool,
                self.data.input_variables[0].node_id,
                self.data.input_variables[0].variable_name,
            )
            if isinstance(history, list):
                history.extend(messages)
                messages = history
        if self.data.stream:
            stream = await litellm.acompletion(
                model=self.data.model.name,
                messages=messages,
                stream=True,
                api_key=get_api_key(self.data.model.name),
                **self.data.model.completion_params
            )

            full_response = ""
            async for chunk in stream:
                content = chunk.choices[0].delta.content
                if content:
                    full_response += content
                    yield NodeRunStreamEvent(
                        node_id=self.data.id,
                        node_title=self.data.title,
                        node_type=self.data.type.value,
                        chunk=content,
                        is_print=self.data.print_stream,
                    )
        else:
            response = litellm.completion(
                model=self.data.model.name,
                messages=messages,
                api_key=get_api_key(self.data.model.name),
                **self.data.model.completion_params
            )
            full_response = response.choices[0].message.content

        yield NodeRunCompletedEvent(
            node_id=self.data.id,
            node_title=self.data.title,
            node_type=self.data.type.value,
            outputs={"llm_output": full_response},
            complete_output=self.data.complete_output,
            is_print=self.data.print_complete,
        )

    def prepare_messages(self, variable_pool):
        messages = []
        for message in self.data.prompt_template:
            content = self.replace_variables(message["content"], variable_pool)
            messages.append({"role": message["role"], "content": content})
        return messages


def get_api_key(model_name: str) -> str:
    if model_name.find("TAIDE-") != -1:
        return os.getenv("TAIDE_API_KEY", "")
    elif model_name.startswith("anthropic"):
        return os.getenv("ANTHROPIC_API_KEY", "")
    elif model_name.startswith("openai"):
        return os.getenv("OPENAI_API_KEY", "")
    elif model_name.startswith("gemini"):
        return os.getenv("GEMINI_API_KEY", "")
    elif model_name.startswith("together_ai"):
        return os.getenv("TOGETHERAI_API_KEY", "")
    elif model_name.startswith("deepinfra"):
        return os.getenv("DEEPINFRA_API_KEY", "")
    else:
        return ""



from fastapi import APIRouter, HTTPException, Depends, Body
from dotenv import load_dotenv
from google.oauth2 import service_account

import os
from pathlib import Path

from botrun_flow_lang.models.botrun_app import BotrunApp
from botrun_flow_lang.models.variable import InputVariable, OutputVariable
from botrun_flow_lang.models.workflow_config import WorkflowConfig
from botrun_flow_lang.models.nodes.llm_node import LLMModelConfig, LLMNodeData
from botrun_flow_lang.models.nodes.start_node import StartNode, StartNodeData
from botrun_flow_lang.models.nodes.end_node import EndNodeData
from botrun_flow_lang.models.workflow import WorkflowData, Workflow
from botrun_flow_lang.api.workflow.workflow_engine import run_workflow
from pydantic import BaseModel, Field

# Include hatch routes
from botrun_flow_lang.api import hatch_api
from botrun_flow_lang.api import user_setting_api
from botrun_flow_lang.api import search_api
from botrun_flow_lang.api import user_workflow_api

load_dotenv()

router = APIRouter(prefix="/api")
router.include_router(
    hatch_api.router,
)
router.include_router(
    user_setting_api.router,
)
router.include_router(
    search_api.router,
)
router.include_router(
    user_workflow_api.router,
)


@router.get("/hello")
async def hello():
    env_name = os.getenv("ENV_NAME")
    google_service_account_key_path = os.getenv(
        "GOOGLE_APPLICATION_CREDENTIALS_FOR_FASTAPI",
        "/app/keys/scoop-386004-d22d99a7afd9.json",
    )
    credentials = service_account.Credentials.from_service_account_file(
        google_service_account_key_path,
        scopes=["https://www.googleapis.com/auth/datastore"],
    )

    return {"message": f"Hello World {env_name}"}


@router.get("/flow/{id}")
async def get_flow(id: str):
    botrun_app = BotrunApp(name="波文件問答", description="給波文件問答用的app")

    start_node = StartNodeData(
        title="Start",
    )

    model_config = LLMModelConfig(
        completion_params={
            "max_tokens": 4096,
            "temperature": 0.7,
        },
        mode="chat",
        name="gpt-4o-2024-08-06",
        provider="openai",
    )
    llm_node = LLMNodeData(
        title="LLM",
        model=model_config,
        prompt_template=[
            {
                "role": "system",
                "content": "妳是臺灣人，回答要用臺灣繁體中文正式用語，需要的時候也可以用英文，可以親切、俏皮、幽默，但不能隨便輕浮。在使用者合理的要求下請盡量配合他的需求，不要隨便拒絕",
            },
            {
                "role": "user",
                "content": f"{{{{#{start_node.id}.user_input#}}}}",
            },
        ],
        input_variables=[
            InputVariable(node_id=start_node.id, variable_name="user_input")
        ],
        output_variables=[
            OutputVariable(variable_name="llm_output"),
        ],
    )
    end_node = EndNodeData(
        title="End",
        input_variables=[
            InputVariable(node_id=llm_node.id, variable_name="llm_output")
        ],
    )
    workflow = WorkflowData(nodes=[start_node, llm_node, end_node])
    original_workflow_config = WorkflowConfig(botrun_app=botrun_app, workflow=workflow)
    yaml_str = original_workflow_config.to_yaml()

    # parent_dir = Path(__file__).parent.parent

    # yaml_file_path = parent_dir / "templates" / f"chatbot_template.yml"
    # with open(yaml_file_path, "w", encoding="utf-8") as yaml_file:
    #     yaml_file.write(yaml_str)

    return {
        "message": f"Flow with id: {id}",
        "yaml": yaml_str,
        # "yaml_file_path": str(yaml_file_path),
    }


class FlowExecutionRequest(BaseModel):
    flow_id: str
    user_input: str


class FlowExecutionResponse(BaseModel):
    message: str
    final_output: str


@router.post("/flow", response_model=FlowExecutionResponse)
async def execute_flow(request: FlowExecutionRequest = Body(...)):
    # 获取工作流配置
    response = await get_flow(request.flow_id)
    yaml_str = response.get("yaml")

    if not yaml_str:
        raise HTTPException(status_code=404, detail="Flow not found")

    # 从 YAML 字符串直接创建 Workflow 对象
    workflow_config = WorkflowConfig.from_yaml(yaml_str)
    workflow = Workflow.from_workflow_data(workflow_config.workflow)

    initial_variable_pool = {}
    # 设置 StartNode 的 user_input
    for node in workflow.nodes:
        if isinstance(node, StartNode):
            initial_variable_pool[node.data.id] = {}
            initial_variable_pool[node.data.id][
                node.data.output_variables[0].variable_name
            ] = request.user_input
            break

    variable_pool = await run_workflow(workflow, initial_variable_pool)

    # 获取最终输出
    final_output = variable_pool.get(workflow.nodes[-1].data.id, {}).get(
        workflow.nodes[-1].data.output_variables[0].variable_name, "No output generated"
    )

    return FlowExecutionResponse(
        message=f"Executed flow with id: {request.flow_id}", final_output=final_output
    )
